import unittest
import copy

from portable_spreadsheet.grammar_utils import GrammarUtils
from portable_spreadsheet.grammars import GRAMMARS


class TestGrammarUtils(unittest.TestCase):
    """Tests of the class GrammarUtils"""

    def test_get_languages(self):
        """Test the getter for system languages."""
        expected = set([str(lang) for lang in GRAMMARS.keys()])
        computed = GrammarUtils.get_languages()
        self.assertSetEqual(expected, computed)

    def test_validate_grammar(self):
        """Test the validator of the grammar."""
        wrong_new_grammar = {'test': 'value'}
        with self.assertRaises(ValueError):
            GrammarUtils.validate_grammar(wrong_new_grammar, True)
        self.assertFalse(GrammarUtils.validate_grammar(wrong_new_grammar,
                                                       False))

        correct_grammar = copy.deepcopy(GRAMMARS['native'])
        self.assertTrue(GrammarUtils.validate_grammar(correct_grammar, True))

    def test_add_remove_grammar(self):
        """Tests if adding and removing of grammars works."""
        # Test if exception is raised when incorrect grammar is added
        wrong_new_grammar = {'test': 'value'}
        with self.assertRaises(ValueError):
            GrammarUtils.validate_grammar(wrong_new_grammar, "wrong")
        # Test adding and removing
        grammars_before: set = copy.deepcopy(GrammarUtils.get_languages())
        new_grammar = GRAMMARS['native']
        language_for_new_grammar = "newly-added"
        # Test adding
        GrammarUtils.add_grammar(new_grammar, language_for_new_grammar)
        grammars_after = copy.deepcopy(grammars_before)
        grammars_after.add(language_for_new_grammar)
        set_after_adding = GrammarUtils.get_languages()
        self.assertSetEqual(set_after_adding, grammars_after)
        # Testing removing
        GrammarUtils.remove_grammar(language_for_new_grammar)
        set_after_deleting = GrammarUtils.get_languages()
        self.assertSetEqual(set_after_deleting, grammars_before)

    def test_system_consistency(self):
        """Check the method for probing system consistency."""
        self.assertTrue(GrammarUtils.check_system_consistency())



from typing import Dict, Optional, Iterable, Tuple, Callable

import numpy as np
import numpy_financial as npf

from .word_constructor import WordConstructor
from .cell_type import CellType
from .cell_indices import CellIndices


class Cell(object):
    """Represent the single cell in the spreadsheet.

    Attributes:
        row (Optional[int]): Position of the cell (row index).
        column (Optional[int]): Position of the cell (column index).
        cell_type (CellType): If the cell is Value or Computational. Value type
            means that it only stores constant value, computation means that
            it does some computations
        _value (Optional[float]): The actual value of the cell.
        cell_indices (CellIndices): The indices of the columns and rows for
            each used language.
        _constructing_words (WordConstructor): The words defining the cell in
            each language.
        is_variable (bool): If True, cell is considered to be a varaible.
        variable_name (Optional[str]): The name of variable.
        _excel_format (dict): Dictionary defining the Excel format style
            for the cell.
        _description (Optional[str]): Optional description of the cell.
    """
    def __init__(self,
                 row: Optional[int] = None,
                 column: Optional[int] = None, /,  # noqa E999
                 value: Optional[float] = None, *,
                 cell_indices: CellIndices,
                 # Private arguments
                 cell_type: CellType = CellType.value_only,
                 words: Optional[WordConstructor] = None,
                 is_variable: bool = False,
                 variable_name: Optional[str] = None,
                 ):
        """Create single cell.

        Args:
            row (Optional[int]): Position of the cell (row index).
            column (Optional[int]): Position of the cell (column index).
            cell_indices (CellIndices): The indices of the columns and rows for
                each used language.
            cell_type (CellType): If the cell is Value or Computational. Value
                type means that it only stores constant value, computation
                means that it does some computations. Do not use this argument
                directly.
            words (Optional[WordConstructor]): The words defining the cell
                in each language. Do not use this argument directly.
            is_variable (bool): If True, cell is considered to be a variable.
            variable_name (Optional[str]): The name of variable.
        """
        if row is not None and column is None \
                or row is None and column is not None:
            raise ValueError("The values of 'row' and 'column' parameters have"
                             " to be either both not None or both None.")
        self.row: Optional[int] = row
        self.column: Optional[int] = column
        self._value: Optional[float] = value
        self.cell_type: CellType = cell_type
        self.cell_indices: cell_indices = cell_indices  # pass only reference
        self.is_variable: bool = is_variable
        self.variable_name: Optional[str] = variable_name
        self._excel_format: dict = {}
        self._description: Optional[str] = None

        if words is not None:
            self._constructing_words: WordConstructor = words
        else:
            self._constructing_words: WordConstructor = \
                WordConstructor.init_from_new_cell(self)

    # === CLASS METHODS and PROPERTIES: ===
    @property
    def word(self) -> WordConstructor:
        """Return correct word. If the cell is anchored, returns just the
            coordinates. If the cell is not anchored: 1) If it is computational
            returns the building string, 2) If it is Value returns the value

        Returns:
              WordConstructor: Right word.
        """
        if self.anchored:
            return WordConstructor.reference(self)
        else:
            if self.cell_type == CellType.computational:
                return self._constructing_words
            elif self.cell_type == CellType.value_only:
                return WordConstructor.constant(self)

    @property
    def anchored(self) -> bool:
        """Return True if the cell is the part of the grid. Return False
            if it is temporarily created for the computation.

        Returns:
            bool: True/False value telling if the cell is anchored to grid.
        """
        return self.row is not None

    @property
    def coordinates(self) -> Tuple[int, int]:
        """Returns the coordinates of the cell.

        Returns:
            Tuple[int, int]: 1) row index, 2) column index
        """
        return self.row, self.column

    @coordinates.setter
    def coordinates(self, coordinates: Tuple[int, int]) -> None:
        """Set the coordinates of the grid.

        Args:
            coordinates (Tuple[int, int]): New coordinates index (row, column)
        """
        self.row = coordinates[0]
        self.column = coordinates[1]

    @property
    def value(self) -> float:
        """Return the actual numeric value of the cell.

        Returns:
            float: the numeric value of the cell.
        """
        return self._value

    def __str__(self) -> str:
        """Override the method __str__ to allow human-readable string
            generations of inside.

        Returns:
            str: String representation of inner value.
        """
        return str(self.parse)

    @property
    def constructing_words(self) -> WordConstructor:
        """Get the constructing words.

        Returns:
            WordConstructor: the constructing word of the cell.
        """
        return self._constructing_words

    @property
    def parse(self) -> Dict[str, str]:
        """Return the dictionary with keys: language, values: constructing
            word. This function is called when the cell should be inserted to
            a spreadsheet.

        Returns:
            Dict[str, str]: Words for each language
        """
        return self._constructing_words.parse(self)

    @property
    def excel_format(self) -> dict:
        """Return style/format of the cell for Excel.

        Returns:
            dict: Excel format dictionary
        """
        return self._excel_format

    @excel_format.setter
    def excel_format(self, new_format: dict):
        """Set the Excel cell format/style.

        Read the documentation: https://xlsxwriter.readthedocs.io/format.html

        Args:
            new_format (dict): New format definition.
        """
        if not isinstance(new_format, dict):
            raise ValueError("New format has to be a dictionary!")
        self._excel_format = new_format

    @property
    def description(self) -> Optional[str]:
        """Return description of the cell or None.

        Returns:
            str: description of the cell or None.
        """
        return self._description

    @description.setter
    def description(self, new_description: Optional[str]):
        """Set the cell description.

        Args:
            new_description (Optional[str]): description of the cell.
        """
        # Use only None value or string value
        if (new_description is not None
                and not isinstance(new_description, str)):
            raise ValueError("Cell description has to be a string value!")
        self._description = new_description
    # =====================================

    # === BINARY OPERATORS: ===
    def add(self, other: 'Cell', /) -> 'Cell':  # noqa: E225
        """Add two values.

        Args:
            other (Cell): Another operand.

        Returns:
            Cell: self + other
        """
        return Cell(value=self.value + other.value,
                    words=WordConstructor.add(self, other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    def subtract(self, other: 'Cell', /) -> 'Cell':  # noqa: E225
        """Subtract two values.

        Args:
            other (Cell): Another operand.

        Returns:
            Cell: self - other
        """
        return Cell(value=self.value - other.value,
                    words=WordConstructor.subtract(self, other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    def multiply(self, other: 'Cell', /) -> 'Cell':  # noqa: E225
        """Multiply two values.

        Args:
            other (Cell): Another operand.

        Returns:
            Cell: self * other
        """
        return Cell(value=self.value * other.value,
                    words=WordConstructor.multiply(self, other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    def divide(self, other: 'Cell', /) -> 'Cell':  # noqa: E225
        """Divide two values.

        Args:
            other (Cell): Another operand.

        Returns:
            Cell: self / other
        """
        return Cell(value=self.value / other.value,
                    words=WordConstructor.divide(self, other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    def modulo(self, other: 'Cell', /) -> 'Cell':  # noqa: E225
        """Modulo of two values.

        Args:
            other (Cell): Another operand.

        Returns:
            Cell: self / other
        """
        return Cell(value=self.value % other.value,
                    words=WordConstructor.modulo(self, other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    def power(self, other: 'Cell', /) -> 'Cell':  # noqa: E225
        """Self power to other.

        Args:
            other (Cell): Another operand.

        Returns:
            Cell: self ** other
        """
        return Cell(value=self.value ** other.value,
                    words=WordConstructor.power(self, other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    def equalTo(self, other: 'Cell', /) -> 'Cell':  # noqa: E225
        """Boolean equal operator.

        Args:
            other (Cell): Another operand.

        Returns:
            Cell: self equal to other
        """
        return Cell(value=self.value == other.value,
                    words=WordConstructor.equalTo(self, other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    def notEqualTo(self, other: 'Cell', /) -> 'Cell':  # noqa: E225
        """Boolean not equal operator.

        Args:
            other (Cell): Another operand.

        Returns:
            Cell: self not equal to other
        """
        return Cell(value=self.value != other.value,
                    words=WordConstructor.notEqualTo(self, other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    def greaterThan(self, other: 'Cell', /) -> 'Cell':  # noqa: E225
        """Boolean greater than operator.

        Args:
            other (Cell): Another operand.

        Returns:
            Cell: self greater than other
        """
        return Cell(value=self.value > other.value,
                    words=WordConstructor.greaterThan(self, other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    def greaterThanOrEqualTo(self, other: 'Cell', /) -> 'Cell':  # noqa: E225
        """Boolean greater than or equal to operator.

        Args:
            other (Cell): Another operand.

        Returns:
            Cell: self greater than or equal to other
        """
        return Cell(value=self.value >= other.value,
                    words=WordConstructor.greaterThanOrEqualTo(self, other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    def lessThan(self, other: 'Cell', /) -> 'Cell':  # noqa: E225
        """Boolean less than operator.

        Args:
            other (Cell): Another operand.

        Returns:
            Cell: self less than other
        """
        return Cell(value=self.value < other.value,
                    words=WordConstructor.lessThan(self, other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    def lessThanOrEqualTo(self, other: 'Cell', /) -> 'Cell':  # noqa: E225
        """Boolean less than or equal to operator.

        Args:
            other (Cell): Another operand.

        Returns:
            Cell: self less than or equal to other
        """
        return Cell(value=self.value <= other.value,
                    words=WordConstructor.lessThanOrEqualTo(self, other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    def logicalConjunction(self, other: 'Cell', /) -> 'Cell':  # noqa: E225
        """Logical conjunction operator.

        Args:
            other (Cell): Another operand.

        Returns:
            Cell: self and other
        """
        return Cell(value=self.value and other.value,
                    words=WordConstructor.logicalConjunction(self, other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    def logicalDisjunction(self, other: 'Cell', /) -> 'Cell':  # noqa: E225
        """Logical disjunction operator.

        Args:
            other (Cell): Another operand.

        Returns:
            Cell: self or other
        """
        return Cell(value=self.value or other.value,
                    words=WordConstructor.logicalDisjunction(self, other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    def concatenate(self, other: 'Cell', /) -> 'Cell':  # noqa: E225
        """Concatenate two values as strings.

        Args:
            other (Cell): Another operand.

        Returns:
            Cell: self concatenate with other
        """
        return Cell(value=str(self.value) + str(other.value),
                    words=WordConstructor.concatenate(self, other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )
    # =========================

    # ==== OPERATOR OVERLOADING ====
    def __mul__(self, other):
        """Overload the operator '*'."""
        return self.multiply(other)

    def __sub__(self, other):
        """Overload the operator '-'."""
        return self.subtract(other)

    def __add__(self, other):
        """Overload the operator '+'."""
        return self.add(other)

    def __truediv__(self, other):
        """Overload the operator '/'."""
        return self.divide(other)

    def __pow__(self, power, modulo=None):
        """Overload the operator '**'."""
        return self.power(power)

    def __mod__(self, other):
        """Overload the operator '%'."""
        return self.modulo(other)

    def __eq__(self, other):
        """Overload the operator '=='."""
        return self.equalTo(other)

    def __ne__(self, other):
        """Overload the operator '!='."""
        return self.notEqualTo(other)

    def __gt__(self, other):
        """Overload the operator '>'."""
        return self.greaterThan(other)

    def __ge__(self, other):
        """Overload the operator '>='."""
        return self.greaterThanOrEqualTo(other)

    def __lt__(self, other):
        """Overload the operator '<'."""
        return self.lessThan(other)

    def __le__(self, other):
        """Overload the operator '<='."""
        return self.lessThanOrEqualTo(other)

    def __and__(self, other):
        """Overload the operator 'and'."""
        return self.logicalConjunction(other)

    def __or__(self, other):
        """Overload the operator 'or'."""
        return self.logicalDisjunction(other)

    def __invert__(self):
        """Overload the operator '~'."""
        return self.logicalNegation(self)

    def __lshift__(self, other):
        """Overload the operator '<<' to string concatenation."""
        return self.concatenate(other)
    # ==============================

    # === AGGREGATE OPERATORS ===
    @staticmethod
    def sum(cell_start: 'Cell', cell_end: 'Cell',
            subset: Iterable['Cell']) -> 'Cell':
        """Compute the sum of the slice.

        Args:
            cell_start (Cell): Starting cell of the slice (left top).
            cell_end: (Cell'): Ending cell of the slice (bottom right).
            subset (Iterable['Cell']): List of all cells in the subset.

        Returns:
            Cell: the cell with aggregated and computed results.
        """
        return Cell._aggregate_fun(cell_start, cell_end, subset,
                                   'sum', np.sum)

    @staticmethod
    def product(cell_start: 'Cell', cell_end: 'Cell',
                subset: Iterable['Cell']) -> 'Cell':
        """Compute the product of the slice.

        Args:
            cell_start (Cell): Starting cell of the slice (left top).
            cell_end: (Cell'): Ending cell of the slice (bottom right).
            subset (Iterable['Cell']): List of all cells in the subset.

        Returns:
            Cell: the cell with aggregated and computed results.
        """
        return Cell._aggregate_fun(cell_start, cell_end, subset,
                                   'product', np.prod)

    @staticmethod
    def mean(cell_start: 'Cell', cell_end: 'Cell',
             subset: Iterable['Cell']) -> 'Cell':
        """Compute the mean-average of the slice.

        Args:
            cell_start (Cell): Starting cell of the slice (left top).
            cell_end: (Cell'): Ending cell of the slice (bottom right).
            subset (Iterable['Cell']): List of all cells in the subset.

        Returns:
            Cell: the cell with aggregated and computed results.
        """
        return Cell._aggregate_fun(cell_start, cell_end, subset,
                                   'mean', np.mean)

    @staticmethod
    def min(cell_start: 'Cell', cell_end: 'Cell',
            subset: Iterable['Cell']) -> 'Cell':
        """Compute the minimum of the slice.

        Args:
            cell_start (Cell): Starting cell of the slice (left top).
            cell_end: (Cell'): Ending cell of the slice (bottom right).
            subset (Iterable['Cell']): List of all cells in the subset.

        Returns:
            Cell: the cell with aggregated and computed results.
        """
        return Cell._aggregate_fun(cell_start, cell_end, subset,
                                   'minimum', np.min)

    @staticmethod
    def max(cell_start: 'Cell', cell_end: 'Cell',
            subset: Iterable['Cell']) -> 'Cell':
        """Compute the maximum of the slice.

        Args:
            cell_start (Cell): Starting cell of the slice (left top).
            cell_end: (Cell'): Ending cell of the slice (bottom right).
            subset (Iterable['Cell']): List of all cells in the subset.

        Returns:
            Cell: the cell with aggregated and computed results.
        """
        return Cell._aggregate_fun(cell_start, cell_end, subset,
                                   'maximum', np.max)

    @staticmethod
    def stdev(cell_start: 'Cell', cell_end: 'Cell',
              subset: Iterable['Cell']) -> 'Cell':
        """Compute the standard deviation of the slice.

        Args:
            cell_start (Cell): Starting cell of the slice (left top).
            cell_end: (Cell'): Ending cell of the slice (bottom right).
            subset (Iterable['Cell']): List of all cells in the subset.

        Returns:
            Cell: the cell with aggregated and computed results.
        """
        return Cell._aggregate_fun(cell_start, cell_end, subset,
                                   'stdev', np.std)

    @staticmethod
    def median(cell_start: 'Cell', cell_end: 'Cell',
               subset: Iterable['Cell']) -> 'Cell':
        """Compute the median of the slice.

        Args:
            cell_start (Cell): Starting cell of the slice (left top).
            cell_end: (Cell'): Ending cell of the slice (bottom right).
            subset (Iterable['Cell']): List of all cells in the subset.

        Returns:
            Cell: the cell with aggregated and computed results.
        """
        return Cell._aggregate_fun(cell_start, cell_end, subset,
                                   'median', np.median)

    @staticmethod
    def count(cell_start: 'Cell', cell_end: 'Cell',
              subset: Iterable['Cell']) -> 'Cell':
        """Compute the number of items the slice.

        Args:
            cell_start (Cell): Starting cell of the slice (left top).
            cell_end: (Cell'): Ending cell of the slice (bottom right).
            subset (Iterable['Cell']): List of all cells in the subset.

        Returns:
            Cell: the cell with aggregated and computed results.
        """
        return Cell._aggregate_fun(cell_start, cell_end, subset,
                                   'count', len)

    @staticmethod
    def irr(cell_start: 'Cell', cell_end: 'Cell',
            subset: Iterable['Cell']) -> 'Cell':
        """Compute the Internal Rate of Return (IRR) of items in slice.

        Args:
            cell_start (Cell): Starting cell of the slice (left top).
            cell_end: (Cell'): Ending cell of the slice (bottom right).
            subset (Iterable['Cell']): List of all cells in the subset.

        Returns:
            Cell: return the Internal Rate of Return (IRR) of slice.
        """
        return Cell._aggregate_fun(cell_start, cell_end, subset,
                                   'irr', npf.irr)

    @staticmethod
    def match_negative_before_positive(cell_start: 'Cell', cell_end: 'Cell',
                                       subset: Iterable['Cell']) -> 'Cell':
        """Find the position of the last negative number in the series that
            is located just before the first non-negative number.

        Args:
            cell_start (Cell): Starting cell of the slice (left top).
            cell_end: (Cell'): Ending cell of the slice (bottom right).
            subset (Iterable['Cell']): List of all cells in the subset.

        Returns:
            Cell: Return the position of the negative number in a series that
                is located just before the first positive number (or zero).
        """
        return Cell._aggregate_fun(cell_start, cell_end, subset,
                                   'match-negative-before-positive',
                                   lambda x: np.argmin(np.array(x) < 0)
                                   )

    @staticmethod
    def _aggregate_fun(
            cell_start: 'Cell',
            cell_end: 'Cell',
            subset: Iterable['Cell'],
            grammar_method: str,
            method_np: Callable[[Iterable[float]], float]
    ) -> 'Cell':
        """General aggregation function covering all methods.

        Args:
            cell_start (Cell): Starting cell of the slice (left top).
            cell_end: (Cell'): Ending cell of the slice (bottom right).
            subset (Iterable['Cell']): List of all cells in the subset.
            grammar_method (str): What method is used (like 'minimum', ...)
            method_np (Callable[[Iterable[float]], float]): What numpy method
                is used for computation.

        Returns:
            Cell: the cell with aggregated and computed results.

        Raises:
            ValueError: If the starting or ending cells are not anchored.
        """
        if not(cell_start.anchored and cell_end.anchored):
            raise ValueError("All cells in the slice has to be anchored!")

        return Cell(value=method_np([c.value for c in subset]),
                    words=WordConstructor.aggregation(
                        cell_start, cell_end, grammar_method
                    ),
                    cell_indices=cell_start.cell_indices,
                    cell_type=CellType.computational
                    )
    # ===========================

    # === UNARY OPERATORS: ===
    @staticmethod
    def reference(other: 'Cell', /) -> 'Cell':  # noqa E225
        """Create a reference to some anchored cell.

        Args:
            other (Cell): The cell to what reference is constructed.

        Returns:
            Cell: Cell containing only the reference to another cell.
        """
        if not other.anchored:
            raise ValueError("The referenced cell has to be anchored.")

        return Cell(value=other.value,
                    words=WordConstructor.reference(other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    @staticmethod
    def raw(other: 'Cell', words: Dict[str, str], /) -> 'Cell':  # noqa: E225
        """Add the raw statement and use the value of input cell.

        Args:
            other (Cell): Input cell that defines value and type of output.
            words (Dict[str, str]): Word for each language (language is a key,
                word is a value)

        Returns:
            Cell: Expression with defined word
        """
        return Cell(value=other.value,
                    words=WordConstructor.raw(other, words),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    @staticmethod
    def variable(other: 'Cell', /) -> 'Cell':  # noqa E225
        """The cell as a variable.

        Args:
            other (Cell): Cell that is interpreted as a variable.

        Returns:
            Cell: variable.
        """
        if not other.is_variable:
            raise ValueError("Only the variable type cell is accepted!")
        return Cell(value=other.value,
                    words=WordConstructor.variable(other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    @staticmethod
    def brackets(other: 'Cell', /) -> 'Cell':  # noqa E225
        """Add brackets to expression.

        Args:
            other (Cell): The expression that should be in brackets.

        Returns:
            Cell: Expression in brackets
        """
        return Cell(value=other.value,
                    words=WordConstructor.brackets(other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    @staticmethod
    def logarithm(other: 'Cell', /) -> 'Cell':  # noqa E225
        """Logarithm of the value in the cell.

        Args:
            other (Cell): Argument of the logarithm function.

        Returns:
            Cell: logarithm of the value
        """
        return Cell(value=np.log(other.value),
                    words=WordConstructor.logarithm(other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    @staticmethod
    def exponential(other: 'Cell', /) -> 'Cell':  # noqa E225
        """Exponential of the value in the cell.

        Args:
            other (Cell): Argument of the exponential function.

        Returns:
            Cell: exponential of the value
        """
        return Cell(value=np.exp(other.value),
                    words=WordConstructor.exponential(other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    @staticmethod
    def ceil(other: 'Cell', /) -> 'Cell':  # noqa E225
        """Ceiling function of the value in the cell.

        Args:
            other (Cell): Argument of the ceiling function.

        Returns:
            Cell: ceiling function value of the input
        """
        return Cell(value=np.ceil(other.value),
                    words=WordConstructor.ceil(other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    @staticmethod
    def floor(other: 'Cell', /) -> 'Cell':  # noqa E225
        """Floor function of the value in the cell.

        Args:
            other (Cell): Argument of the floor function.

        Returns:
            Cell: floor function value of the input
        """
        return Cell(value=np.floor(other.value),
                    words=WordConstructor.floor(other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    @staticmethod
    def round(other: 'Cell', /) -> 'Cell':  # noqa E225
        """Round the value in the cell.

        Args:
            other (Cell): Argument of the rounding function.

        Returns:
            Cell: round of the input numeric value
        """
        return Cell(value=np.round(other.value),
                    words=WordConstructor.round(other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    @staticmethod
    def abs(other: 'Cell', /) -> 'Cell':  # noqa E225
        """Absolute value of the value in the cell.

        Args:
            other (Cell): Argument of the absolute value function.

        Returns:
            Cell: absolute value of the input numeric value
        """
        return Cell(value=np.abs(other.value),
                    words=WordConstructor.abs(other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    @staticmethod
    def sqrt(other: 'Cell', /) -> 'Cell':  # noqa E225
        """Square root of the value in the cell.

        Args:
            other (Cell): Argument of the square root function.

        Returns:
            Cell: square root of the input numeric value
        """
        return Cell(value=np.sqrt(other.value),
                    words=WordConstructor.sqrt(other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    @staticmethod
    def signum(other: 'Cell', /) -> 'Cell':  # noqa E225
        """Signum function of the value in the cell.

        Args:
            other (Cell): Argument of the signum function.

        Returns:
            Cell: signum of the input numeric value
        """
        return Cell(value=np.sign(other.value),
                    words=WordConstructor.signum(other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )

    @staticmethod
    def logicalNegation(other: 'Cell', /) -> 'Cell':  # noqa E225
        """Logical negation of the value in the cell.

        Args:
            other (Cell): Argument of the logical negation.

        Returns:
            Cell: logical negation the input value
        """
        return Cell(value=not other.value,
                    words=WordConstructor.logicalNegation(other),
                    cell_indices=other.cell_indices,
                    cell_type=CellType.computational
                    )
    # ========================

    # === Conditional (if-then-else statement) ===
    @staticmethod
    def conditional(condition: 'Cell',
                    consequent: 'Cell',
                    alternative: 'Cell', /) -> 'Cell':  # noqa E225
        """Conditional statement (standard if-then-else statement).

        Evaluate the value of the condition, if it is true, take the value
            of the consequent, if not take the value of the alternative.

        Args:
            condition (Cell): Cell defining the condition (boolean value)
            consequent (Cell): What is pass if the condition is true (the
                part right after the if)
            alternative (Cell): What is pass if the condition is false (the
                part right after the else)

        Returns:
            Cell: The if-then-else conditional statement and correctly
                chosen value.
        """
        return Cell(value=(consequent.value
                           if condition.value else alternative.value),
                    words=WordConstructor.conditional(
                        condition, consequent, alternative
                    ),
                    cell_indices=condition.cell_indices,
                    cell_type=CellType.computational
                    )
    # ============================================

    # === OFFSET ===
    @staticmethod
    def offset(reference: 'Cell',
               row_skip: 'Cell',
               column_skip: 'Cell', /, *, # noqa E225
               target: 'Cell') -> 'Cell':
        """Return the cell with value computed as offset from reference cell
            plus row_skip rows and column_skip columns.

        Args:
            reference (Cell): Reference cell from that the position is
                computed.
            row_skip (Cell): How many rows (down) should be skipped.
            column_skip (Cell): How many columns (left) should be skipped.
            target (Cell): Target cell (place where user really skips).

        Returns:
            Cell: Cell with offset operation and value from the cell on the
                referential position.

        Raises:
            ValueError: If any of reference or target cells are not anchored.
        """
        # Test if both reference and target cells are anchored
        if not(reference.anchored and target.anchored):
            raise ValueError("Both reference cell and target cell must be"
                             " anchored!")
        return Cell(value=target.value,
                    words=WordConstructor.offset(
                        reference, row_skip, column_skip
                    ),
                    cell_indices=reference.cell_indices,
                    cell_type=CellType.computational
                    )
    # ==============



import unittest

import copy

import numpy as np

from portable_spreadsheet.spreadsheet import Spreadsheet
from portable_spreadsheet.cell import Cell
from portable_spreadsheet.cell_type import CellType
from portable_spreadsheet.cell_slice import CellSlice
from portable_spreadsheet.cell_indices import CellIndices
from portable_spreadsheet.cell_indices_templates import excel_generator


class TestSpreadsheetBasicFunctionality(unittest.TestCase):
    """Test the basic spreadsheet basic functionality."""
    def setUp(self) -> None:
        self.warnings = []
        self.nr_row = 20
        self.nr_col = 30
        self.rows_labels = [f"R_{r_i}" for r_i in range(self.nr_row)]
        self.columns_labels = [f"NL_C_{c_i}" for c_i in range(self.nr_col)]
        self.rows_help_text = [f"HT_R_{r_i}" for r_i in range(self.nr_row)]
        self.columns_help_text = [f"HT_C_{c_i}" for c_i in range(self.nr_col)]
        self.native_rows = [f"NL_R_{r_i}" for r_i in range(self.nr_row)]
        self.native_cols = [f"NL_C_{c_i}" for c_i in range(self.nr_col)]
        self.sheet = Spreadsheet.create_new_sheet(
            self.nr_row, self.nr_col, {
                'native': (
                    self.native_rows,
                    self.native_cols
                )
            },
            rows_labels=self.rows_labels,
            columns_labels=self.columns_labels,
            rows_help_text=self.rows_help_text,
            columns_help_text=self.columns_help_text,
            excel_append_row_labels=True,
            excel_append_column_labels=True,
            warning_logger=lambda message: self.warnings.append(message)
        )
        self.sheet_shape = (self.nr_row, self.nr_col)

    def test_index_property(self):
        """Test index property"""
        computed = self.sheet.index
        expected = self.rows_labels
        self.assertListEqual(expected, computed)
        # Slice
        computed = self.sheet.iloc[3:, :].index
        expected = self.rows_labels[3:]
        self.assertListEqual(expected, computed)

    def test_columns_property(self):
        """Test columns property"""
        computed = self.sheet.columns
        expected = self.columns_labels
        self.assertListEqual(expected, computed)
        # Slice
        computed = self.sheet.iloc[:, 3:].columns
        expected = self.columns_labels[3:]
        self.assertListEqual(expected, computed)

    def test_create_new_sheet(self):
        """Test the instance sheet"""
        self.assertTrue(isinstance(self.sheet, Spreadsheet))

    def test_expand_sheet(self):
        """Test the expanding of the sheet size"""
        # Try to expand without all parameters
        with self.assertRaises(ValueError):
            self.sheet.cell_indices.expand_size(5, 3)

        old_sheet: Spreadsheet = copy.deepcopy(self.sheet)
        expand_row = 3
        expand_col = 5
        new_native_rows = [f'eR_{r_i}' for r_i in range(expand_row)]
        new_native_cols = [f'eC_{c_i}' for c_i in range(expand_col)]
        new_rows_labels = [f'LeR_{r_i}' for r_i in range(expand_row)]
        new_columns_labels = [f'LeC_{c_i}' for c_i in range(expand_col)]
        new_rows_help_text = [f'HeR_{r_i}' for r_i in range(expand_row)]
        new_columns_help_text = [f'HeC_{c_i}' for c_i in range(expand_col)]

        new_cell_idx = self.sheet.cell_indices.expand_size(
            expand_row, expand_col,
            {
                'native': (
                    new_native_rows,
                    new_native_cols
                )
            },
            new_rows_labels=new_rows_labels,
            new_columns_labels=new_columns_labels,
            new_rows_help_text=new_rows_help_text,
            new_columns_help_text=new_columns_help_text
        )
        old_sheet.expand_using_cell_indices(new_cell_idx)
        # Test the expanded sheet
        cell_indices: CellIndices = old_sheet.cell_indices
        self.assertTupleEqual(
            new_cell_idx.shape,
            old_sheet.shape,
            (self.nr_row + expand_row, self.nr_col + expand_col)
        )
        # Check rows in each language
        self.assertListEqual(cell_indices.rows['native'],
                             self.native_rows + new_native_rows)
        self.assertListEqual(
            cell_indices.rows['python_numpy'],
            [str(i) for i in range(0, self.nr_row + expand_row + 1)]
        )
        self.assertListEqual(
            cell_indices.rows['excel'],
            [str(i) for i in range(2, self.nr_row + expand_row + 1 + 1)]
        )
        # Check columns in each language
        self.assertListEqual(cell_indices.columns['native'],
                             self.native_cols + new_native_cols)
        self.assertListEqual(
            cell_indices.columns['python_numpy'],
            [str(i) for i in range(0, self.nr_col + expand_col + 1)]
        )
        self.assertListEqual(
            cell_indices.columns['excel'],
            excel_generator(1, self.nr_col + expand_col, 1, 1)[1]
        )
        # Check variables of cell index
        self.assertListEqual(cell_indices.rows_labels,
                             self.rows_labels + new_rows_labels)
        self.assertListEqual(cell_indices.columns_labels,
                             self.columns_labels + new_columns_labels)
        self.assertListEqual(cell_indices.rows_help_text,
                             self.rows_help_text + new_rows_help_text)
        self.assertListEqual(cell_indices.columns_help_text,
                             self.columns_help_text + new_columns_help_text)

    def test_cell_indices(self):
        """Test the indices inside sheet"""
        cell_indices: CellIndices = self.sheet.cell_indices
        self.assertTupleEqual(cell_indices.shape, self.sheet_shape)
        # Check rows in each language
        self.assertListEqual(cell_indices.rows['native'], self.native_rows)
        self.assertListEqual(cell_indices.rows['python_numpy'],
                             [str(i) for i in range(0, self.nr_row + 1)])
        self.assertListEqual(cell_indices.rows['excel'],
                             [str(i) for i in range(2, self.nr_row + 1 + 1)])
        # Check columns in each language
        self.assertListEqual(cell_indices.columns['native'],
                             self.native_cols)
        self.assertListEqual(cell_indices.columns['python_numpy'],
                             [str(i) for i in range(0, self.nr_col + 1)])
        self.assertListEqual(cell_indices.columns['excel'],
                             excel_generator(1, self.nr_col, 1, 1)[1])
        self.assertListEqual(excel_generator(1, 5, 1, 1)[1],
                             ["B", "C", "D", "E", "F"])
        # Check variables of cell index
        self.assertListEqual(cell_indices.rows_labels, self.rows_labels)
        self.assertListEqual(cell_indices.columns_labels, self.columns_labels)
        self.assertListEqual(cell_indices.rows_help_text, self.rows_help_text)
        self.assertListEqual(cell_indices.columns_help_text,
                             self.columns_help_text)

    def test_shape_property(self):
        """Check the shape of the sheet property"""
        self.assertTupleEqual(self.sheet.shape, self.sheet_shape)

    def test_warning(self):
        """Test the warning logs."""
        self.sheet.iloc[:, :].to_dictionary()
        self.assertEqual(len(self.warnings), 1)


class TestSpreadsheetSelection(unittest.TestCase):
    """Test the selecting (slicing) from the spreadsheet."""
    def setUp(self) -> None:
        self.n_row = 13
        self.n_col = 27
        self.row_labels = [f"R_{r_i}" for r_i in range(self.n_row)]
        self.col_labels = [f"C_{c_i}" for c_i in range(self.n_col)]
        self.sheet = Spreadsheet.create_new_sheet(
            self.n_row, self.n_col,
            rows_labels=self.row_labels,
            columns_labels=self.col_labels
        )

    def assertAllClose2D(self, operand_1, operand_2, delta=0.000001) -> None:
        """Compare two 2D numpy-like arrays (with shape property)

        Args:
            operand_1: Expected value
            operand_2: Compared value
            delta: Delta for IEEE-Double comparisons
        """
        self.assertTupleEqual(operand_1.shape, operand_2.shape)
        for row_idx in range(operand_1.shape[0]):
            for col_idx in range(operand_1.shape[1]):
                if (np.isnan(operand_1[row_idx, col_idx])
                        and np.isnan(operand_2[row_idx, col_idx])):
                    continue
                self.assertAlmostEqual(operand_1[row_idx, col_idx],
                                       operand_2[row_idx, col_idx],
                                       delta=delta)

    def test_single_values(self):
        """Test the selecting and writing to the single value"""
        sheet = copy.deepcopy(self.sheet)
        np_sheet: np.ndarray = np.full((self.n_row, self.n_col), np.nan)
        # Getting values out of range
        with self.assertRaises(IndexError):
            self.sheet.iloc[50, 1]
        with self.assertRaises(IndexError):
            self.sheet.iloc[1, 55]

        # Setting values out of range
        with self.assertRaises(IndexError):
            self.sheet.iloc[50, 1] = 7
        with self.assertRaises(IndexError):
            self.sheet.iloc[1, 55] = 7

        # Getting values out of by label
        with self.assertRaises(ValueError):
            self.sheet.loc["XYZ", "C_1"]
        with self.assertRaises(ValueError):
            self.sheet.loc["R_1", "XZY"]

        # Setting values out of range by label
        with self.assertRaises(ValueError):
            self.sheet.loc["XYZ", "C_1"] = 7
        with self.assertRaises(ValueError):
            self.sheet.loc["R_1", "XZY"] = 7

        # Set the single value
        i_idx = (11, 20)
        sheet.iloc[i_idx] = 89.912
        np_sheet[i_idx] = 89.912
        self.assertAllClose2D(sheet.to_numpy(), np_sheet)
        # Test getter
        self.assertAlmostEqual(sheet.iloc[i_idx].value, np_sheet[i_idx])
        # Test the word created in cell
        a_cell: Cell = sheet.iloc[i_idx]
        self.assertEqual(a_cell.word.words['python_numpy'],
                         f'values[{i_idx[0]},{i_idx[1]}]')
        self.assertTrue(a_cell.anchored)
        self.assertEqual(a_cell.cell_type, CellType.value_only)
        self.assertTupleEqual(a_cell.coordinates, i_idx)

        # Setting by the label:
        sheet.loc['R_11', "C_20"] = 7.3
        np_sheet[i_idx] = 7.3
        self.assertAllClose2D(sheet.to_numpy(), np_sheet)
        # Test getter
        self.assertAlmostEqual(sheet.iloc[i_idx].value, np_sheet[i_idx])
        # Test the word created in cell
        a_cell: Cell = sheet.iloc[i_idx]
        self.assertEqual(a_cell.word.words['python_numpy'],
                         f'values[{i_idx[0]},{i_idx[1]}]')
        self.assertTrue(a_cell.anchored)
        self.assertEqual(a_cell.cell_type, CellType.value_only)
        self.assertTupleEqual(a_cell.coordinates, i_idx)

        # Test negative indices
        sheet.iloc[-7, -11] = 35.123
        np_sheet[-7, -11] = 35.123
        self.assertAllClose2D(sheet.to_numpy(), np_sheet)
        # Test getter
        self.assertAlmostEqual(sheet.iloc[-7, -11].value, np_sheet[-7, -11])

    def test_slice(self):
        """Test the selecting and writing to the slice"""
        sheet = copy.deepcopy(self.sheet)
        np_sheet: np.ndarray = np.full((self.n_row, self.n_col), np.nan)
        # Getting values out of range
        with self.assertRaises(IndexError):
            self.sheet.iloc[:500, 1]
        with self.assertRaises(IndexError):
            self.sheet.iloc[1, :500]

        # Setting values out of range
        with self.assertRaises(IndexError):
            self.sheet.iloc[500:, 1] = 7
        with self.assertRaises(IndexError):
            self.sheet.iloc[1, 500:] = 7

        # Getting values out of range by label
        with self.assertRaises(ValueError):
            self.sheet.loc[:"R_600", "C_1"]
        with self.assertRaises(ValueError):
            self.sheet.loc["R_1", :"C_30000"]

        # Setting values out of range by label
        with self.assertRaises(ValueError):
            self.sheet.loc[:"R_600", "C_1"] = 7
        with self.assertRaises(ValueError):
            self.sheet.loc["R_1", :"C_30000"] = 7

        # Set the slice of values
        i_idx = (slice(0, 11), slice(0, 20))
        sheet.iloc[i_idx] = 89.912
        np_sheet[i_idx] = 89.912
        self.assertAllClose2D(sheet.to_numpy(), np_sheet)
        # Test getter
        self.assertAllClose2D(sheet.iloc[i_idx].to_numpy(), np_sheet[i_idx])
        # Test slice method
        cell_slice: CellSlice = sheet.iloc[i_idx]
        u_max = cell_slice.max()
        self.assertDictEqual(u_max.word.words,
                             {'python_numpy': 'np.max(values[0:11,0:20])',
                              'excel': 'MAX(B2:U12)'})
        self.assertFalse(u_max.anchored)
        self.assertEqual(u_max.cell_type, CellType.computational)
        self.assertTupleEqual(u_max.coordinates, (None, None))
        # Test slice setter
        cell_slice <<= 9.9
        np_sheet[i_idx] = 9.9
        self.assertAllClose2D(sheet.iloc[i_idx].to_numpy(), np_sheet[i_idx])

        # Set the slice of values using labels
        label_idx = (slice("R_0", "R_11"), slice("C_0", "C_20"))
        sheet.loc[label_idx] = 3.14159
        np_sheet[i_idx] = 3.14159
        # Test getter
        self.assertAllClose2D(sheet.loc[label_idx].to_numpy(), np_sheet[i_idx])

        # Test slice with negative values:
        # Set the slice of values
        i_idx = (slice(0, -3), slice(0, -8))
        sheet.iloc[i_idx] = 14.5458
        np_sheet[i_idx] = 14.5458
        self.assertAllClose2D(sheet.to_numpy(), np_sheet)
        # Test getter
        self.assertAllClose2D(sheet.iloc[i_idx].to_numpy(), np_sheet[i_idx])



from typing import List, Tuple


def general_interface(rows: int,
                      columns: int,
                      offset_row: int = 0,
                      offset_column: int = 0) -> Tuple[List[str], List[str]]:
    """Generate the series of the labels for rows and columns.

    Args:
        rows (int): How many rows should be generated
        columns (int): How many columns should be generated
        offset_row (int) The row offset from the beginning
        offset_column (int) The column offset from the beginning

    Return:
        Tuple[List[str], List[str]]: Tuple of indices for rows and columns
    """
    # This is the template interface
    pass


def excel_generator(rows: int,
                    columns: int,
                    offset_row: int = 0,
                    offset_column: int = 0) -> Tuple[List[str], List[str]]:
    rows = list([str(row)
                 for row in range(1 + offset_row, rows + 1 + offset_row)])
    letter_list = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    cols = []
    for column in range(offset_column, columns + offset_column):
        col = column + 1
        result = []
        while col:
            col, rem = divmod(col - 1, 26)
            result[:0] = letter_list[rem]
        cols.append(''.join(result))
    return rows, cols


def python_numpy_generator(rows: int,
                           columns: int,
                           offset_row: int = 0,
                           offset_column: int = 0
                           ) -> Tuple[List[str], List[str]]:
    # the + 1 value is because of offset for slices
    rows = list([str(row) for row in range(offset_row, rows + offset_row + 1)])
    # the + 1 value is because of offset for slices
    cols = list([str(col) for col in range(offset_column,
                                           columns + offset_column + 1)])
    return rows, cols


def native_generator(rows: int,
                     columns: int,
                     offset_row: int = 0,
                     offset_column: int = 0) -> Tuple[List[str], List[str]]:
    rows = list([str(row) for row in range(
        1 + offset_row, rows + offset_row + 1)]
                )
    cols = list([str(col) for col in range(
        1 + offset_column, columns + offset_column + 1)])
    return rows, cols


cell_indices_generators = {
    "excel": excel_generator,
    "python_numpy": python_numpy_generator,
    "native": native_generator,
}
system_languages = ['excel', 'python_numpy']



from typing import List, Tuple, Dict, Optional, Callable
import copy

from .grammars import GRAMMARS
from .cell_indices_templates import cell_indices_generators, system_languages

# ==== TYPES ====
# mapping from language to list, used for mapping from language to list rows
#   names or column names
T_lg_ar = Dict[str, List[str]]
T_lg_col_row = Dict[str, Tuple[List[str], List[str]]]
# ===============


class CellIndices(object):
    """Represent the indices of the cells and its labels for each language.
    """
    def __init__(self,
                 number_of_rows: int,
                 number_of_columns: int,
                 rows_columns: Optional[T_lg_col_row] = None,
                 /, *,  # noqa E999
                 rows_labels: List[str] = None,
                 columns_labels: List[str] = None,
                 rows_help_text: List[str] = None,
                 columns_help_text: List[str] = None,
                 excel_append_row_labels: bool = True,
                 excel_append_column_labels: bool = True,
                 warning_logger: Optional[Callable[[str], None]] = None
                 ):
        """Create cell indices object.

        Args:
            number_of_rows (int): Number of rows.
            number_of_columns (int): Number of columns.
            rows_columns (T_lg_col_row): List of all row names and column names
                for each language.
            rows_labels (List[str]): List of masks (aliases) for row
                names.
            columns_labels (List[str]): List of masks (aliases) for column
                names.
            rows_help_text (List[str]): List of help texts for each row.
            columns_help_text (List[str]): List of help texts for each column.
            excel_append_row_labels (bool): If True, one column is added
                on the beginning of the sheet as a offset for labels.
            excel_append_column_labels (bool): If True, one row is added
                on the beginning of the sheet as a offset for labels.
            warning_logger (Optional[Callable[[str], None]]): Function that
                logs the warnings (or None if skipped).
        """
        # Quick sanity check:
        if rows_labels is not None \
                and len(set(rows_labels)) != len(rows_labels):
            warning_logger("There are some duplications in row labels.")
        if columns_labels is not None \
                and len(set(columns_labels)) != len(columns_labels):
            warning_logger("There are some duplications in column labels.")

        if rows_columns is not None:
            for language in rows_columns.keys():
                # Does the language include the last cell?
                #   if yes, offset of size 1 has to be included.
                offset = 0
                if GRAMMARS[
                    language
                ]['cells']['aggregation']['include_last_cell']:
                    offset = 1
                rows, columns = rows_columns[language]
                if len(rows) != number_of_rows + offset:
                    e_mess = "Number of rows is not the same for every" \
                             "language! Or you have not included offset " \
                             "caused by excluding the last value of the slice!"
                    raise ValueError(e_mess)
                if len(columns) != number_of_columns + offset:
                    e_mess = "Number of columns is not the same for every" \
                             "language! Or you have not included offset " \
                             "caused by excluding the last value of the slice!"
                    raise ValueError(e_mess)
        if number_of_rows < 1 or number_of_columns < 1:
            raise ValueError("Number of rows and columns has to at least 1!")
        # check the columns and rows aliases sizes
        if (rows_labels is not None
                and len(rows_labels) != number_of_rows):
            raise ValueError("Number of rows aliases has to be the same"
                             "as number of rows!")
        if (columns_labels is not None
                and len(columns_labels) != number_of_columns):
            raise ValueError("Number of columns aliases has to be the same"
                             "as number of columns!")
        # check the help texts sizes
        if (rows_help_text is not None
                and len(rows_help_text) != number_of_rows):
            raise ValueError("Number of rows help texts has to be the same"
                             "as number of rows!")
        if (columns_help_text is not None
                and len(columns_help_text) != number_of_columns):
            raise ValueError("Number of columns help texts has to be the same"
                             "as number of columns!")
        # -------------------
        self.number_of_rows: int = number_of_rows
        self.number_of_columns: int = number_of_columns
        self.excel_append_row_labels: bool = excel_append_row_labels
        self.excel_append_column_labels: bool = excel_append_column_labels
        self.rows: T_lg_ar = {}
        self.columns: T_lg_ar = {}
        # Append the system languages
        for language, generator in cell_indices_generators.items():
            if language not in system_languages:
                continue
            offset_row = 0
            offset_column = 0
            if self.excel_append_row_labels and language == "excel":
                offset_column = 1
            if self.excel_append_column_labels and language == "excel":
                offset_row = 1

            rows, cols = generator(self.number_of_rows,
                                   self.number_of_columns,
                                   offset_row,
                                   offset_column)
            self.rows[language] = rows
            self.columns[language] = cols
        # Append the not-system languages and user defined languages
        self.user_defined_languages: List[str] = []
        if rows_columns is not None:
            for language, values in rows_columns.items():
                rows, cols = values
                self.rows[language] = rows
                self.columns[language] = cols
                self.user_defined_languages.append(language)
        # Define user defined names for rows and columns
        self.rows_labels: List[str] = copy.deepcopy(rows_labels)
        self.columns_labels: List[str] = copy.deepcopy(columns_labels)
        # Or define auto generated aliases as an integer sequence from 0
        if rows_labels is None:
            self.rows_labels = [str(_row_n) for _row_n in
                                range(number_of_rows)]
        if columns_labels is None:
            self.columns_labels = [str(_col_n) for _col_n in
                                   range(number_of_columns)]
        # assign the help texts
        self.rows_help_text: List[str] = copy.deepcopy(rows_help_text)
        self.columns_help_text: List[str] = copy.deepcopy(columns_help_text)

    @property
    def supported_languages(self) -> List[str]:
        """Returns all languages supported by the indicies.

        Returns:
            List[str]: All languages supported by this indices.
        """
        return [].extend(self.user_defined_languages, system_languages)

    @property
    def shape(self) -> Tuple[int]:
        """Return the shape of the object in the NumPy logic.

        Returns:
            Tuple[int]: Number of rows, Number of columns
        """
        language = list(self.columns.keys())[0]
        return len(self.rows[language]), len(self.columns[language])

    @property
    def languages(self) -> List[str]:
        """Return all supported languages.

        Returns:
            List[str]: List of all supported languages.
        """
        return [str(lan) for lan in self.rows.keys()]

    def expand_size(self,
                    new_number_of_rows: int,
                    new_number_of_columns: int,
                    new_rows_columns: Optional[T_lg_col_row] = {},
                    /, *,  # noqa E225
                    new_rows_labels: List[str] = None,
                    new_columns_labels: List[str] = None,
                    new_rows_help_text: List[str] = None,
                    new_columns_help_text: List[str] = None,
                    ) -> 'CellIndices':
        """Expand the size of the table.

        Args:
            new_number_of_rows (int): Number of rows to be added.
            new_number_of_columns (int): Number of columns to be added.
            new_rows_columns (T_lg_col_row): List of all row names and column
                names for each language to be added.
            new_rows_labels (List[str]): List of masks (aliases) for row
                names to be added.
            new_columns_labels (List[str]): List of masks (aliases) for
                column names to be added.
            new_rows_help_text (List[str]): List of help texts for each row to
                be added.
            new_columns_help_text (List[str]): List of help texts for each
                column to be added.
        """
        expanded = copy.deepcopy(self)
        # Quick sanity check:
        if new_number_of_rows < 1 and new_number_of_columns < 1:
            return
        # check the columns and rows aliases sizes
        if (new_rows_labels is not None
                and len(new_rows_labels) != new_number_of_rows):
            raise ValueError("Number of rows aliases has to be the same"
                             "as number of rows!")
        if (new_columns_labels is not None
                and len(new_columns_labels) != new_number_of_columns):
            raise ValueError("Number of columns aliases has to be the same"
                             "as number of columns!")
        # check the help texts sizes
        if (new_rows_help_text is not None
                and len(new_rows_help_text) != new_number_of_rows):
            raise ValueError("Number of rows help texts has to be the same"
                             "as number of rows!")
        if (new_columns_help_text is not None
                and len(new_columns_help_text) != new_number_of_columns):
            raise ValueError("Number of columns help texts has to be the same"
                             "as number of columns!")
        # -------------------
        # Append the system languages
        for language, generator in cell_indices_generators.items():
            if language not in system_languages:
                continue
            offset_row = 0
            offset_column = 0
            if expanded.excel_append_row_labels and language == "excel":
                offset_column = 1
            if expanded.excel_append_column_labels and language == "excel":
                offset_row = 1
            rows, cols = generator(expanded.number_of_rows +
                                   new_number_of_rows + offset_row,
                                   expanded.number_of_columns +
                                   new_number_of_columns + offset_column)
            expanded.rows[language] = rows[offset_row:]
            expanded.columns[language] = cols[offset_column:]
        # Append rows to user defined languages
        for language, values in new_rows_columns.items():
            # Does the language include the last cell?
            #   if yes, offset of size 1 has to be included.
            offset = 0
            if GRAMMARS[
                language
            ]['cells']['aggregation']['include_last_cell']:
                offset = 1
            rows, cols = values
            # Quick sanity check
            if language not in expanded.user_defined_languages:
                raise ValueError("Users languages has to match to existing "
                                 "ones!")
            if len(rows) != new_number_of_rows + offset:
                raise ValueError("Number of rows has to be the same "
                                 "as number of rows to be added! Plus the"
                                 "offset for the ending row.")
            if len(cols) != new_number_of_columns + offset:
                raise ValueError("Number of columns has to be the same "
                                 "as number of columns to be added! Plus the"
                                 "offset for the ending row.")
            # ------------------
            if offset == 1:
                del expanded.rows[language][-1]
                del expanded.columns[language][-1]

            expanded.rows[language].extend(rows)
            expanded.columns[language].extend(cols)

        # Define user defined names for rows and columns
        if new_rows_labels is not None:
            expanded.rows_labels.extend(new_rows_labels)
        else:
            expanded.rows_labels = [
                str(i)
                for i in range(expanded.number_of_rows + new_number_of_rows)
            ]
        if new_columns_labels is not None:
            expanded.columns_labels.extend(new_columns_labels)
        else:
            expanded.columns_labels = [
                str(i)
                for i in range(
                    expanded.number_of_columns + new_number_of_columns
                )
            ]
        # Or define auto generated aliases as an integer sequence from 0
        if new_columns_labels is None:
            expanded.columns_labels = [
                str(i)
                for i in range(
                    expanded.number_of_columns + new_number_of_columns
                )
            ]
        # assign the help texts
        if expanded.rows_help_text is not None and new_number_of_rows > 0:
            if new_rows_help_text is None:
                raise ValueError("Rows help texts has to set.")
            expanded.rows_help_text.extend(new_rows_help_text)
        if expanded.columns_help_text is not None \
                and new_number_of_columns > 0:
            if new_columns_help_text is None:
                raise ValueError("Columns help texts has to set.")
            expanded.columns_help_text.extend(new_columns_help_text)

        # Modify the number of rows/columns
        expanded.number_of_rows += new_number_of_rows
        expanded.number_of_columns += new_number_of_columns

        # Return new expanded indices
        return expanded



EXCEL = {
    "rows": {
        # like 1, 2, 3...
        "name-regexp": "[0-9]+",
        "maximal-number": 65535
    },
    "cols": {
        # like A, B, C...
        "name-regexp": "[A-Z]+",
        "maximal-number": 65535
    },
    "cells": {
        # like 7
        "constant": {
            "prefix": "",
            "suffix": "",
        },
        "variable": {
            "prefix": "",
            "suffix": "",
            "value": {
                'include': False,
                "prefix": "",
                "suffix": ""
            }
        },
        "empty": {
            "content": ""
        },
        # Like A1, A2
        "reference": {
            "prefix": "",
            "suffix": "",
            "separator": "",
            "row_first": False
        },
        # like =7+A2
        "operation": {
            "prefix": "=",
            "suffix": "",
        },
        # like A1:A50
        "aggregation": {
            "prefix": "",
            "separator": ":",
            "suffix": "",
            "start_cell": {
                "prefix": "",
                "suffix": "",
                "separator": "",
                "rows_only": False,
                "cols_only": False,
                "row_first": False,
            },
            "end_cell": {
                "prefix": "",
                "suffix": "",
                "separator": "",
                "rows_only": False,
                "cols_only": False,
                "row_first": False,
            },
            # For example, Python exclude last cell when slicing
            "include_last_cell": False
        },
        "offset": {
            "order": ('reference-cell-column', 'reference-cell-row',
                      'skip-of-rows', 'skip-of-columns'),
            "prefix": "OFFSET(",
            "suffix": ")",
            'reference-cell-column': {
                "prefix": "",
                "suffix": "",
            },
            'reference-cell-row': {
                "prefix": "",
                "suffix": ",",
            },
            'skip-of-rows': {
                "prefix": "",
                "suffix": ",",
            },
            'skip-of-columns': {
                "prefix": "",
                "suffix": "",
            },
        }
    },
    "operations": {
        # BASIC OPERATIONS
        # like 7+3
        "add": {
            "prefix": "",
            "suffix": "",
            "separator": "+",
        },
        # like 7-3
        "subtract": {
            "prefix": "",
            "suffix": "",
            "separator": "-",
        },
        # like 7*3
        "multiply": {
            "prefix": "",
            "suffix": "",
            "separator": "*",
        },
        # like 7/3
        "divide": {
            "prefix": "",
            "suffix": "",
            "separator": "/",
        },
        # like 7%3
        "modulo": {
            "prefix": "MOD(",
            "suffix": ")",
            "separator": ",",
        },
        # like 7^3
        "power": {
            "prefix": "",
            "suffix": "",
            "separator": "^",
        },
        # Merge two strings
        'concatenate': {
            "prefix": "CONCATENATE(",
            "suffix": ")",
            "separator": ",",
            # Context of string
            "string-value": {
                "prefix": '"',
                "suffix": '"',
            },
            # Context of numbers
            "numeric-value": {
                "prefix": '',
                "suffix": '',
            },
        },
        # AGGREGATE FUNCTIONS
        "sum": {
            "prefix": "SUM(",
            "suffix": ")",
        },
        "product": {
            "prefix": "PRODUCT(",
            "suffix": ")",
        },
        "mean": {
            "prefix": "AVERAGE(",
            "suffix": ")",
        },
        "minimum": {
            "prefix": "MIN(",
            "suffix": ")",
        },
        "maximum": {
            "prefix": "MAX(",
            "suffix": ")",
        },
        "stdev": {
            "prefix": "STDEV(",
            "suffix": ")",
        },
        "median": {
            "prefix": "MEDIAN(",
            "suffix": ")",
        },
        "count": {
            "prefix": "COUNT(",
            "suffix": ")",
        },
        "irr": {
            "prefix": "IRR(",
            "suffix": ")",
        },
        "match-negative-before-positive": {
            "prefix": "MATCH(0,",
            "suffix": ")",
        },
        # BASIC OPERATIONS
        "exponential": {
            "prefix": "EXP(",
            "suffix": ")",
        },
        "logarithm": {
            "prefix": "LN(",
            "suffix": ")",
        },
        "ceil": {
            "prefix": "CEILING(",
            "suffix": ")",
        },
        "floor": {
            "prefix": "FLOOR(",
            "suffix": ")",
        },
        "round": {
            "prefix": "ROUND(",
            "suffix": ")",
        },
        "abs": {
            "prefix": "ABS(",
            "suffix": ")",
        },
        "sqrt": {
            "prefix": "SQRT(",
            "suffix": ")",
        },
        "signum": {
            "prefix": "SIGN(",
            "suffix": ")",
        },
        # LOGICAL OPERATIONS
        "equal-to": {
            "prefix": "",
            "suffix": "",
            "separator": "=",
        },
        "not-equal-to": {
            "prefix": "",
            "suffix": "",
            "separator": "<>",
        },
        "greater-than": {
            "prefix": "",
            "suffix": "",
            "separator": ">",
        },
        "greater-than-or-equal-to": {
            "prefix": "",
            "suffix": "",
            "separator": ">=",
        },
        "less-than": {
            "prefix": "",
            "suffix": "",
            "separator": "<",
        },
        "less-than-or-equal-to": {
            "prefix": "",
            "suffix": "",
            "separator": "<=",
        },
        "logical-conjunction": {
            "prefix": "AND(",
            "suffix": ")",
            "separator": ", ",
        },
        "logical-disjunction": {
            "prefix": "OR(",
            "suffix": ")",
            "separator": ", ",
        },
        "logical-negation": {
            "prefix": "NOT(",
            "suffix": ")",
        },
    },
    "brackets": {
        "prefix": "(",
        "suffix": ")"
    },
    "conditional": {
        "order": ('condition', 'consequent', 'alternative'),
        "prefix": "IF(",
        "suffix": ")",
        'condition': {
            "prefix": "",
            "suffix": ","
        },
        'consequent': {
            "prefix": "",
            "suffix": ","
        },
        'alternative': {
            "prefix": "",
            "suffix": ""
        }
    }
}
NATIVE = {
    "rows": {
        # like date...
        "name-regexp": "[0-9]+",
        "maximal-number": 65535
    },
    "cols": {
        # like discount...
        "name-regexp": "[0-9]+",
        "maximal-number": 65535
    },
    "cells": {
        # like 7
        "constant": {
            "prefix": "",
            "suffix": "",
        },
        "variable": {
            "prefix": "value of variable '",
            "suffix": "'",
            "value": {
                'include': True,
                "prefix": " (=",
                "suffix": ")"
            }
        },
        "empty": {
            "content": ""
        },
        # Like value at (1.1.2015, 25%)
        "reference": {
            "prefix": "value at (",
            "suffix": ")",
            "separator": ", ",
            "row_first": True
        },
        # like 7 + value at (1.1.2015, 25%)
        "operation": {
            "prefix": "",
            "suffix": "",
        },
        # like values from (1.1.2015, 25%) to (1.1.2015, 50%)
        "aggregation": {
            "prefix": "values from ",
            "separator": " and ",
            "suffix": "",
            "start_cell": {
                "prefix": "row ",
                "suffix": "",
                "separator": " to row ",
                "rows_only": True,
                "cols_only": False,
                "row_first": True,
            },
            "end_cell": {
                "prefix": "from column ",
                "suffix": "",
                "separator": " to column ",
                "rows_only": False,
                "cols_only": True,
                "row_first": True,
            },
            # For example, Python exclude last cell when slicing
            "include_last_cell": False
        },
        "offset": {
            "order": ('reference-cell-row', 'skip-of-rows',
                      'reference-cell-column', 'skip-of-columns'),
            "prefix": "skip from ",
            "suffix": "",
            'reference-cell-column': {
                "prefix": "and from the column at position ",
                "suffix": " exactly ",
            },
            'reference-cell-row': {
                "prefix": "the row at position ",
                "suffix": " exactly ",
            },
            'skip-of-rows': {
                "prefix": "",
                "suffix": " items down, ",
            },
            'skip-of-columns': {
                "prefix": "",
                "suffix": " items left",
            },
        }
    },
    "operations": {
        # BASIC OPERATIONS
        # like 7+3
        "add": {
            "prefix": "",
            "suffix": "",
            "separator": " + ",
        },
        # like 7 - 3
        "subtract": {
            "prefix": "",
            "suffix": "",
            "separator": " - ",
        },
        # like 7*3
        "multiply": {
            "prefix": "",
            "suffix": "",
            "separator": " * ",
        },
        # like 7/3
        "divide": {
            "prefix": "",
            "suffix": "",
            "separator": " / ",
        },
        # like 7%3
        "modulo": {
            "prefix": "",
            "suffix": "",
            "separator": " mod ",
        },
        # Merge two strings
        'concatenate': {
            "prefix": "concatenate string ",
            "suffix": " and ",
            "separator": "",
            # Context of string
            "string-value": {
                "prefix": "'",
                "suffix": "'",
            },
            # Context of numbers
            "numeric-value": {
                "prefix": "'",
                "suffix": "'",
            },
        },
        # like 7^3
        "power": {
            "prefix": "",
            "suffix": "",
            "separator": " power to ",
        },
        # AGGREGATE FUNCTIONS
        "sum": {
            "prefix": "sum of ",
            "suffix": "",
        },
        "product": {
            "prefix": "product of ",
            "suffix": "",
        },
        "mean": {
            "prefix": "mean-average of ",
            "suffix": "",
        },
        "minimum": {
            "prefix": "minimum of ",
            "suffix": "",
        },
        "maximum": {
            "prefix": "maximum of ",
            "suffix": "",
        },
        "stdev": {
            "prefix": "standard deviation of ",
            "suffix": "",
        },
        "median": {
            "prefix": "median of ",
            "suffix": "",
        },
        "count": {
            "prefix": "number of items in the slice ",
            "suffix": "",
        },
        "irr": {
            "prefix": "Internal Rate of Return of ",
            "suffix": "",
        },
        "match-negative-before-positive": {
            "prefix": "position of the step where numbers "
                      "become to be positive of ",
            "suffix": "",
        },
        # BASIC OPERATIONS
        "exponential": {
            "prefix": "exponential function of ",
            "suffix": "",
        },
        "logarithm": {
            "prefix": "logarithm of ",
            "suffix": "",
        },
        "ceil": {
            "prefix": "ceiling function of ",
            "suffix": "",
        },
        "floor": {
            "prefix": "floor function of ",
            "suffix": "",
        },
        "round": {
            "prefix": "round of ",
            "suffix": "",
        },
        "abs": {
            "prefix": "absolute value of ",
            "suffix": "",
        },
        "sqrt": {
            "prefix": "square root of ",
            "suffix": "",
        },
        "signum": {
            "prefix": "signum of ",
            "suffix": "",
        },
        # LOGICAL OPERATIONS
        "equal-to": {
            "prefix": "",
            "suffix": "",
            "separator": " equal to ",
        },
        "not-equal-to": {
            "prefix": "",
            "suffix": "",
            "separator": " not equal to ",
        },
        "greater-than": {
            "prefix": "",
            "suffix": "",
            "separator": " greater than ",
        },
        "greater-than-or-equal-to": {
            "prefix": "",
            "suffix": "",
            "separator": " greater than or equal to ",
        },
        "less-than": {
            "prefix": "",
            "suffix": "",
            "separator": " less than ",
        },
        "less-than-or-equal-to": {
            "prefix": "",
            "suffix": "",
            "separator": " less than or equal to ",
        },
        "logical-conjunction": {
            "prefix": "",
            "suffix": "",
            "separator": " and ",
        },
        "logical-disjunction": {
            "prefix": "",
            "suffix": "",
            "separator": " or ",
        },
        "logical-negation": {
            "prefix": "negation (",
            "suffix": ")",
        },
    },
    "brackets": {
        "prefix": "(",
        "suffix": ")"
    },
    "conditional": {
        "order": ('condition', 'consequent', 'alternative'),
        "prefix": "(",
        "suffix": ")",
        'condition': {
            "prefix": "if ",
            "suffix": ""
        },
        'consequent': {
            "prefix": " then ",
            "suffix": ""
        },
        'alternative': {
            "prefix": " else ",
            "suffix": ""
        }
    }
}
PYTHON_NUMPY = {
    "rows": {
        # like 0,1,2...
        "name-regexp": "[0-9]+",
        "maximal-number": 65535
    },
    "cols": {
        # like 0,1,2...
        "name-regexp": "[0-9]+",
        "maximal-number": 65535
    },
    "cells": {
        # like 7
        "constant": {
            "prefix": "",
            "suffix": "",
        },
        "empty": {
            "content": ""
        },
        "variable": {
            "prefix": "",
            "suffix": "",
            "value": {
                'include': False,
                "prefix": "",
                "suffix": ""
            }
        },
        # Like [row,column]
        "reference": {
            "prefix": "values[",
            "suffix": "]",
            "separator": ",",
            "row_first": True
        },
        # like 7+values[1,2] (in Python case do nothing)
        "operation": {
            "prefix": "",
            "suffix": "",
        },
        # like values[1:2,3:7]
        "aggregation": {
            "prefix": "values[",
            "separator": ",",
            "suffix": "]",
            "start_cell": {
                "prefix": "",
                "suffix": "",
                "separator": ":",
                "rows_only": True,
                "cols_only": False,
                "row_first": True
            },
            "end_cell": {
                "prefix": "",
                "suffix": "",
                "separator": ":",
                "rows_only": False,
                "cols_only": True,
                "row_first": True
            },
            # For example, Python exclude last cell when slicing
            "include_last_cell": True
        },
        "offset": {
            "order": ('reference-cell-row', 'skip-of-rows',
                      'reference-cell-column', 'skip-of-columns'),
            "prefix": "values[",
            "suffix": "]",
            'reference-cell-column': {
                "prefix": "",
                "suffix": "+",
            },
            'reference-cell-row': {
                "prefix": "",
                "suffix": "+",
            },
            'skip-of-rows': {
                "prefix": "",
                "suffix": ",",
            },
            'skip-of-columns': {
                "prefix": "",
                "suffix": "",
            },
        }
    },
    "operations": {
        # BASIC OPERATIONS
        # like 7+3
        "add": {
            "prefix": "",
            "suffix": "",
            "separator": "+",
        },
        # like 7-3
        "subtract": {
            "prefix": "",
            "suffix": "",
            "separator": "-",
        },
        # like 7*3
        "multiply": {
            "prefix": "",
            "suffix": "",
            "separator": "*",
        },
        # like 7/3
        "divide": {
            "prefix": "",
            "suffix": "",
            "separator": "/",
        },
        # like 7%3
        "modulo": {
            "prefix": "",
            "suffix": "",
            "separator": "%",
        },
        # like 7^3
        "power": {
            "prefix": "",
            "suffix": "",
            "separator": "**",
        },
        # Merge two strings
        'concatenate': {
            "prefix": "",
            "suffix": "",
            "separator": "+",
            # Context of string
            "string-value": {
                "prefix": '',
                "suffix": '',
            },
            # Context of numbers
            "numeric-value": {
                "prefix": '"',
                "suffix": '"',
            },
        },
        # AGGREGATE FUNCTIONS
        "sum": {
            "prefix": "np.sum(",
            "suffix": ")",
        },
        "product": {
            "prefix": "np.prod(",
            "suffix": ")",
        },
        "mean": {
            "prefix": "np.mean(",
            "suffix": ")",
        },
        "minimum": {
            "prefix": "np.min(",
            "suffix": ")",
        },
        "maximum": {
            "prefix": "np.max(",
            "suffix": ")",
        },
        "stdev": {
            "prefix": "np.std(",
            "suffix": ")",
        },
        "median": {
            "prefix": "np.median(",
            "suffix": ")",
        },
        "count": {
            "prefix": "((lambda var=",
            "suffix": ": var.shape[0] * var.shape[1])())",
        },
        "irr": {
            "prefix": "npf.irr(",
            "suffix": ")",
        },
        "match-negative-before-positive": {
            "prefix": "np.argmin(",
            "suffix": "<0)",
        },
        # BASIC OPERATIONS
        "exponential": {
            "prefix": "np.exp(",
            "suffix": ")",
        },
        "logarithm": {
            "prefix": "np.log(",
            "suffix": ")",
        },
        "ceil": {
            "prefix": "np.ceil(",
            "suffix": ")",
        },
        "floor": {
            "prefix": "np.floor(",
            "suffix": ")",
        },
        "round": {
            "prefix": "np.round(",
            "suffix": ")",
        },
        "abs": {
            "prefix": "np.abs(",
            "suffix": ")",
        },
        "sqrt": {
            "prefix": "np.sqrt(",
            "suffix": ")",
        },
        "signum": {
            "prefix": "np.sign(",
            "suffix": ")",
        },
        # LOGICAL OPERATIONS
        "equal-to": {
            "prefix": "",
            "suffix": "",
            "separator": "==",
        },
        "not-equal-to": {
            "prefix": "",
            "suffix": "",
            "separator": "!=",
        },
        "greater-than": {
            "prefix": "",
            "suffix": "",
            "separator": ">",
        },
        "greater-than-or-equal-to": {
            "prefix": "",
            "suffix": "",
            "separator": ">=",
        },
        "less-than": {
            "prefix": "",
            "suffix": "",
            "separator": "<",
        },
        "less-than-or-equal-to": {
            "prefix": "",
            "suffix": "",
            "separator": "<=",
        },
        "logical-conjunction": {
            "prefix": "",
            "suffix": "",
            "separator": " and ",
        },
        "logical-disjunction": {
            "prefix": "",
            "suffix": "",
            "separator": " or ",
        },
        "logical-negation": {
            "prefix": "not (",
            "suffix": ")",
        },
    },
    "brackets": {
        "prefix": "(",
        "suffix": ")"
    },
    "conditional": {
        "order": ('consequent', 'condition', 'alternative'),
        "prefix": "(",
        "suffix": ")",
        'condition': {
            "prefix": " if (",
            "suffix": ") "
        },
        'consequent': {
            "prefix": "(",
            "suffix": ")"
        },
        'alternative': {
            "prefix": "else (",
            "suffix": ")"
        }
    }
}

GRAMMARS = {
    # System grammars:
    "excel": EXCEL,
    "python_numpy": PYTHON_NUMPY,
    # User defined grammars goes here:
    "native": NATIVE,
    # --------------------------
}

"""General pattern of a grammar defining some language and with data types
    of each feature (and with description).

Generally a PREFIX is what is inserted before the word when it is created,
    a SUFFIX is what is inserted after the word when created, a SEPARATOR
    is what is inserted between two words where needed. A CONTENT is what
    is inserted instead of any word.
"""
GRAMMAR_PATTERN: dict = {
    "rows": {
        # Define the RegExp pattern that name of rows should pass
        "name-regexp": str,
        # Define the maximal number of rows
        "maximal-number": int
    },
    "cols": {
        # Define the RegExp pattern that name of colmn should pass
        "name-regexp": str,
        # Define the maximal number of rows
        "maximal-number": int
    },
    "cells": {
        # Define the basic words related to a cell

        # Constant definition (like 7):
        "constant": {
            "prefix": str,
            "suffix": str,
        },

        # Variable definition (e.g. pi)
        "variable": {
            "prefix": str,
            "suffix": str,
            # If is included (parameter include), also add the value inside
            #   the variable as a suffix (always) after variable name
            "value": {
                'include': bool,
                "prefix": str,
                "suffix": str
            }
        },

        # Empty cell definition
        "empty": {
            "content": str
        },

        # Reference to a cell (like =A7 in Excel)
        "reference": {
            "prefix": str,
            "suffix": str,
            "separator": str,
            # If True, the row is on the first position and column on the
            # second position. (e. g. it is False for Excel)
            "row_first": bool
        },

        # Operation (like =CELL + CELL)
        "operation": {
            "prefix": str,
            "suffix": str,
        },

        # Aggregation of cells (e. g. C1:C5 in Excel)
        "aggregation": {
            "prefix": str,
            "separator": str,
            "suffix": str,
            # Word defining the starting cell
            "start_cell": {
                "prefix": str,
                "suffix": str,
                "separator": str,
                # If True, cell contains only reference to rows (mutually
                # exclusive with cols_only.
                "rows_only": bool,
                # If True, cell contains only reference to columns (mutually
                # exclusive with rows_only.
                "cols_only": bool,
                "row_first": bool,
            },
            # Word defining the last cell
            "end_cell": {
                "prefix": str,
                "suffix": str,
                "separator": str,
                # If True, cell contains only reference to rows (mutually
                # exclusive with cols_only.
                "rows_only": bool,
                # If True, cell contains only reference to columns (mutually
                # exclusive with rows_only.
                "cols_only": bool,
                "row_first": bool,
            },
            # For example, Python exclude last cell when slicing (it has a
            # close-left logic), but Excel does include it. If True the last
            # cell is included, if False, the last cell is the previous plus
            # one in the index.
            "include_last_cell": bool
        },
        # Return the reference to the cell n-down top and m-columns left
        "offset": {
            # Some ordered subset of ('reference-cell-row', 'skip-of-rows',
            # 'reference-cell-column', 'skip-of-columns') defining in what
            # order the offset is computed.
            "order": tuple,
            "prefix": str,
            "suffix": str,
            # Label (index) for the column of the reference cell
            'reference-cell-column': {
                "prefix": str,
                "suffix": str,
            },
            # Label (index) for the row of the reference cell
            'reference-cell-row': {
                "prefix": str,
                "suffix": str,
            },
            # How many rows are skipped down (reference to the cell, or value)
            'skip-of-rows': {
                "prefix": str,
                "suffix": str,
            },
            # How many columns are skipped left (reference to cell, or value)
            'skip-of-columns': {
                "prefix": str,
                "suffix": str,
            },
        }
    },
    "operations": {
        # === BASIC BINARY OPERATIONS ===
        # Adding, like 7+3
        "add": {
            "prefix": str,
            "suffix": str,
            "separator": str,  # Typically +
        },
        # Subtracting, like 7-3
        "subtract": {
            "prefix": str,
            "suffix": str,
            "separator": str,  # Typically -
        },
        # Multiplying, like 7*3
        "multiply": {
            "prefix": str,
            "suffix": str,
            "separator": str,  # Typically *
        },
        # Dividing, like 7/3
        "divide": {
            "prefix": str,
            "suffix": str,
            "separator": str,  # Typically /
        },
        # Modulo, like 7%3
        "modulo": {
            "prefix": str,
            "suffix": str,
            "separator": str,  # Typically %
        },
        # Power to, like 7^3
        "power": {
            "prefix": str,
            "suffix": str,
            "separator": str,  # Typically '^' (in Python '**')
        },
        # Merge two strings, like "a" + "b" = "ab"
        'concatenate': {
            "prefix": str,
            "suffix": str,
            "separator": str,
            # Context of the string constant
            "string-value": {
                "prefix": str,
                "suffix": str,
            },
            # Context of the numbers constant
            "numeric-value": {
                "prefix": str,
                "suffix": str,
            },
        },

        # === AGGREGATE FUNCTIONS (functions for aggregated cells) ===
        # Sum of aggregated cells.
        "sum": {
            "prefix": str,  # Like SUM(
            "suffix": str,  # Like )
        },
        # Sum of aggregated cells.
        "product": {
            "prefix": str,  # Like PRODUCT(
            "suffix": str,  # Like )
        },
        # Mean-average of aggregated cells.
        "mean": {
            "prefix": str,  # Like AVERAGE(
            "suffix": str,  # Like )
        },
        # Minimum of aggregated cells.
        "minimum": {
            "prefix": str,  # Like MIN(
            "suffix": str,  # Like )
        },
        # Standard deviation of aggregated cells.
        "stdev": {
            "prefix": str,
            "suffix": str,
        },
        # Median of aggregated cells.
        "median": {
            "prefix": str,
            "suffix": str,
        },
        # Maximum of aggregated cells.
        "maximum": {
            "prefix": str,  # Like MAX(
            "suffix": str,  # Like )
        },
        # Number of aggregated cells.
        "count": {
            "prefix": str,
            "suffix": str,
        },
        # Internal Rate of Return (IRR) of aggregated cells.
        "irr": {
            "prefix": str,
            "suffix": str,
        },
        # Match the position of the last negative number before first
        #   negative number in a sequence (works only for single row/column).
        "match-negative-before-positive": {
            "prefix": str,
            "suffix": str,
        },

        # === BASIC UNARY OPERATIONS (that takes only one operand) ===
        # Exponential (natural) of the argument
        "exponential": {
            "prefix": str,  # Like EXP(
            "suffix": str,  # Like )
        },
        # Natural logarithm of the argument
        "logarithm": {
            "prefix": str,  # Like LN(
            "suffix": str,  # Like )
        },
        # Ceiling function
        "ceil": {
            "prefix": str,
            "suffix": str,
        },
        # Floor function
        "floor": {
            "prefix": str,
            "suffix": str,
        },
        # Round the number
        "round": {
            "prefix": str,
            "suffix": str,
        },
        # Absolute value
        "abs": {
            "prefix": str,
            "suffix": str,
        },
        # Absolute value
        "sqrt": {
            "prefix": str,
            "suffix": str,
        },
        # Signum function (-1 if value is < 0, 0 if val == 0, 1 if val > 0)
        "signum": {
            "prefix": str,
            "suffix": str,
        },

        # === LOGICAL OPERATIONS (returns true or false) ===
        # Equal to
        "equal-to": {
            "prefix": str,
            "suffix": str,
            "separator": str,
        },
        # Not equal to
        "not-equal-to": {
            "prefix": str,
            "suffix": str,
            "separator": str,
        },
        # Greater than
        "greater-than": {
            "prefix": str,
            "suffix": str,
            "separator": str,
        },
        # Greater than or equal to
        "greater-than-or-equal-to": {
            "prefix": str,
            "suffix": str,
            "separator": str,
        },
        # Less than
        "less-than": {
            "prefix": str,
            "suffix": str,
            "separator": str,
        },
        # Less than or equal to
        "less-than-or-equal-to": {
            "prefix": str,
            "suffix": str,
            "separator": str,
        },
        # Logical conjunction
        "logical-conjunction": {
            "prefix": str,
            "suffix": str,
            "separator": str,
        },
        # Logical disjunction
        "logical-disjunction": {
            "prefix": str,
            "suffix": str,
            "separator": str,
        },
        # Logical negation
        "logical-negation": {
            "prefix": str,
            "suffix": str,
        },
    },
    # Brackets around the statement
    "brackets": {
        "prefix": str,  # Almost always just '('
        "suffix": str,  # Almost always just ')'
    },
    # Conditional: standard if-then-else statement
    "conditional": {
        # Defines order of clausules ('consequent', 'condition', 'alternative')
        "order": tuple,
        # Defines prefix of the whole conditional
        "prefix": str,
        # Defines suffix of the whole conditional
        "suffix": str,
        # Conditional statement (e. g. 'A > 5 or C < 5')
        'condition': {
            "prefix": str,
            "suffix": str
        },
        # Consequent (then part): part that occurs if the condition is true.
        'consequent': {
            "prefix": str,
            "suffix": str
        },
        # Alternative (else part): part that occurs if the condition is false.
        'alternative': {
            "prefix": str,
            "suffix": str
        }
    }
}



import unittest
import json

import tempfile
import os
import copy

import numpy as np

from portable_spreadsheet.spreadsheet import Spreadsheet


class TestSerialization(unittest.TestCase):
    """Regression test for serializers."""

    def setUp(self) -> None:
        self.warnings = []
        self.nr_row = 5
        self.nr_col = 4
        self.rows_labels = [f"R_{r_i}" for r_i in range(self.nr_row)]
        self.columns_labels = [f"NL_C_{c_i}" for c_i in range(self.nr_col)]
        self.rows_help_text = [f"HT_R_{r_i}" for r_i in range(self.nr_row)]
        self.columns_help_text = [f"HT_C_{c_i}" for c_i in range(self.nr_col)]
        self.native_rows = [f"NL_R_{r_i}" for r_i in range(self.nr_row)]
        self.native_cols = [f"NL_C_{c_i}" for c_i in range(self.nr_col)]
        self.sheet = Spreadsheet.create_new_sheet(
            self.nr_row, self.nr_col, {
                'native': (
                    self.native_rows,
                    self.native_cols
                )
            },
            rows_labels=self.rows_labels,
            columns_labels=self.columns_labels,
            rows_help_text=self.rows_help_text,
            columns_help_text=self.columns_help_text,
            excel_append_row_labels=True,
            excel_append_column_labels=True,
            warning_logger=lambda message: self.warnings.append(message)
        )
        self.sheet_shape = (self.nr_row, self.nr_col)
        # Add some random values:
        self.inserted_rand_values: np.ndarray = np.array(
            [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12],
             [13, 14, 15, 16], [17, 18, 19, 20]]
        )
        self.sheet.iloc[:, :] = self.inserted_rand_values
        # Add some descriptions:
        for row_idx in range(self.nr_row):
            for col_idx in range(self.nr_col):
                self.sheet.iloc[row_idx, col_idx].description = \
                    f"DescFor{row_idx},{col_idx}"
        # Temporary directory for file exports
        self.working_dir = tempfile.mkdtemp()

    def tearDown(self) -> None:
        """Remove temporary directory content."""
        for file in [
            f for f in os.listdir(self.working_dir)
            if os.path.isfile(os.path.join(self.working_dir, f))
        ]:
            os.remove(os.path.join(self.working_dir, file))

    def test_to_excel(self):
        """Test if the Excel file is created."""
        file_name = "export.xlsx"
        excel_path = os.path.join(self.working_dir, file_name)
        self.sheet.to_excel(excel_path)
        self.assertTrue(os.path.exists(excel_path))

    def test_to_dictionary(self):
        """Test export to dictionary"""
        expected = {'table': {'data': {'rows': {'R_0': {'columns': {'NL_C_0': {'excel': '1', 'python_numpy': '1', 'native': '1', 'value': 1, 'description': 'DescFor0,0', 'column_description': 'HT_C_0'}, 'NL_C_1': {'excel': '2', 'python_numpy': '2', 'native': '2', 'value': 2, 'description': 'DescFor0,1', 'column_description': 'HT_C_1'}, 'NL_C_2': {'excel': '3', 'python_numpy': '3', 'native': '3', 'value': 3, 'description': 'DescFor0,2', 'column_description': 'HT_C_2'}, 'NL_C_3': {'excel': '4', 'python_numpy': '4', 'native': '4', 'value': 4, 'description': 'DescFor0,3', 'column_description': 'HT_C_3'}}, 'row_description': 'HT_R_0'}, 'R_1': {'columns': {'NL_C_0': {'excel': '5', 'python_numpy': '5', 'native': '5', 'value': 5, 'description': 'DescFor1,0', 'column_description': 'HT_C_0'}, 'NL_C_1': {'excel': '6', 'python_numpy': '6', 'native': '6', 'value': 6, 'description': 'DescFor1,1', 'column_description': 'HT_C_1'}, 'NL_C_2': {'excel': '7', 'python_numpy': '7', 'native': '7', 'value': 7, 'description': 'DescFor1,2', 'column_description': 'HT_C_2'}, 'NL_C_3': {'excel': '8', 'python_numpy': '8', 'native': '8', 'value': 8, 'description': 'DescFor1,3', 'column_description': 'HT_C_3'}}, 'row_description': 'HT_R_1'}, 'R_2': {'columns': {'NL_C_0': {'excel': '9', 'python_numpy': '9', 'native': '9', 'value': 9, 'description': 'DescFor2,0', 'column_description': 'HT_C_0'}, 'NL_C_1': {'excel': '10', 'python_numpy': '10', 'native': '10', 'value': 10, 'description': 'DescFor2,1', 'column_description': 'HT_C_1'}, 'NL_C_2': {'excel': '11', 'python_numpy': '11', 'native': '11', 'value': 11, 'description': 'DescFor2,2', 'column_description': 'HT_C_2'}, 'NL_C_3': {'excel': '12', 'python_numpy': '12', 'native': '12', 'value': 12, 'description': 'DescFor2,3', 'column_description': 'HT_C_3'}}, 'row_description': 'HT_R_2'}, 'R_3': {'columns': {'NL_C_0': {'excel': '13', 'python_numpy': '13', 'native': '13', 'value': 13, 'description': 'DescFor3,0', 'column_description': 'HT_C_0'}, 'NL_C_1': {'excel': '14', 'python_numpy': '14', 'native': '14', 'value': 14, 'description': 'DescFor3,1', 'column_description': 'HT_C_1'}, 'NL_C_2': {'excel': '15', 'python_numpy': '15', 'native': '15', 'value': 15, 'description': 'DescFor3,2', 'column_description': 'HT_C_2'}, 'NL_C_3': {'excel': '16', 'python_numpy': '16', 'native': '16', 'value': 16, 'description': 'DescFor3,3', 'column_description': 'HT_C_3'}}, 'row_description': 'HT_R_3'}, 'R_4': {'columns': {'NL_C_0': {'excel': '17', 'python_numpy': '17', 'native': '17', 'value': 17, 'description': 'DescFor4,0', 'column_description': 'HT_C_0'}, 'NL_C_1': {'excel': '18', 'python_numpy': '18', 'native': '18', 'value': 18, 'description': 'DescFor4,1', 'column_description': 'HT_C_1'}, 'NL_C_2': {'excel': '19', 'python_numpy': '19', 'native': '19', 'value': 19, 'description': 'DescFor4,2', 'column_description': 'HT_C_2'}, 'NL_C_3': {'excel': '20', 'python_numpy': '20', 'native': '20', 'value': 20, 'description': 'DescFor4,3', 'column_description': 'HT_C_3'}}, 'row_description': 'HT_R_4'}}}, 'variables': {}, 'row-labels': ['R_0', 'R_1', 'R_2', 'R_3', 'R_4'], 'column-labels': ['NL_C_0', 'NL_C_1', 'NL_C_2', 'NL_C_3']}}  # noqa
        self.assertDictEqual(expected, self.sheet.to_dictionary())

    def test_to_csv(self):
        """Test export to CSV"""
        # Expected value with labels
        expected_no_skip = """Sheet,NL_C_0,NL_C_1,NL_C_2,NL_C_3
R_0,1,2,3,4
R_1,5,6,7,8
R_2,9,10,11,12
R_3,13,14,15,16
R_4,17,18,19,20"""
        # Expected value without labels
        expected_skip = """1,2,3,4
5,6,7,8
9,10,11,12
13,14,15,16
17,18,19,20"""
        self.assertEqual(self.sheet.to_csv(), expected_no_skip)
        self.assertEqual(self.sheet.to_csv(skip_labels=True),
                         expected_skip)

    def test_to_markdown(self):
        """MD (Markdown) language export"""
        expected_no_skip = """| *Sheet* | *NL_C_0* | *NL_C_1* | *NL_C_2* | *NL_C_3* |
|----|----|----|----|----|
| *R_0* | 1 | 2 | 3 | 4 |
| *R_1* | 5 | 6 | 7 | 8 |
| *R_2* | 9 | 10 | 11 | 12 |
| *R_3* | 13 | 14 | 15 | 16 |
| *R_4* | 17 | 18 | 19 | 20 |
"""

        expected_skip = """||||||
|----|----|----|----|
| 1 | 2 | 3 | 4 |
| 5 | 6 | 7 | 8 |
| 9 | 10 | 11 | 12 |
| 13 | 14 | 15 | 16 |
| 17 | 18 | 19 | 20 |
"""
        self.assertEqual(self.sheet.to_markdown(), expected_no_skip)
        self.assertEqual(self.sheet.to_markdown(skip_labels=True),
                         expected_skip)

    def test_to_html_table(self):
        """Test the export to the HTML table"""
        expected_no_skip = '<table><tr><th>Sheet</th><th><a href="javascript:;"  title="HT_C_0">NL_C_0</a></th><th><a href="javascript:;"  title="HT_C_1">NL_C_1</a></th><th><a href="javascript:;"  title="HT_C_2">NL_C_2</a></th><th><a href="javascript:;"  title="HT_C_3">NL_C_3</a></th></tr><tr><td><a href="javascript:;"  title="HT_R_0">R_0</a></td><td><a href="javascript:;"  title="DescFor0,0">1</a></td><td><a href="javascript:;"  title="DescFor0,1">2</a></td><td><a href="javascript:;"  title="DescFor0,2">3</a></td><td><a href="javascript:;"  title="DescFor0,3">4</a></td></tr><tr><td><a href="javascript:;"  title="HT_R_1">R_1</a></td><td><a href="javascript:;"  title="DescFor1,0">5</a></td><td><a href="javascript:;"  title="DescFor1,1">6</a></td><td><a href="javascript:;"  title="DescFor1,2">7</a></td><td><a href="javascript:;"  title="DescFor1,3">8</a></td></tr><tr><td><a href="javascript:;"  title="HT_R_2">R_2</a></td><td><a href="javascript:;"  title="DescFor2,0">9</a></td><td><a href="javascript:;"  title="DescFor2,1">10</a></td><td><a href="javascript:;"  title="DescFor2,2">11</a></td><td><a href="javascript:;"  title="DescFor2,3">12</a></td></tr><tr><td><a href="javascript:;"  title="HT_R_3">R_3</a></td><td><a href="javascript:;"  title="DescFor3,0">13</a></td><td><a href="javascript:;"  title="DescFor3,1">14</a></td><td><a href="javascript:;"  title="DescFor3,2">15</a></td><td><a href="javascript:;"  title="DescFor3,3">16</a></td></tr><tr><td><a href="javascript:;"  title="HT_R_4">R_4</a></td><td><a href="javascript:;"  title="DescFor4,0">17</a></td><td><a href="javascript:;"  title="DescFor4,1">18</a></td><td><a href="javascript:;"  title="DescFor4,2">19</a></td><td><a href="javascript:;"  title="DescFor4,3">20</a></td></tr></table>'  # noqa
        expected_skip = '<table><tr><td><a href="javascript:;"  title="DescFor0,0">1</a></td><td><a href="javascript:;"  title="DescFor0,1">2</a></td><td><a href="javascript:;"  title="DescFor0,2">3</a></td><td><a href="javascript:;"  title="DescFor0,3">4</a></td></tr><tr><td><a href="javascript:;"  title="DescFor1,0">5</a></td><td><a href="javascript:;"  title="DescFor1,1">6</a></td><td><a href="javascript:;"  title="DescFor1,2">7</a></td><td><a href="javascript:;"  title="DescFor1,3">8</a></td></tr><tr><td><a href="javascript:;"  title="DescFor2,0">9</a></td><td><a href="javascript:;"  title="DescFor2,1">10</a></td><td><a href="javascript:;"  title="DescFor2,2">11</a></td><td><a href="javascript:;"  title="DescFor2,3">12</a></td></tr><tr><td><a href="javascript:;"  title="DescFor3,0">13</a></td><td><a href="javascript:;"  title="DescFor3,1">14</a></td><td><a href="javascript:;"  title="DescFor3,2">15</a></td><td><a href="javascript:;"  title="DescFor3,3">16</a></td></tr><tr><td><a href="javascript:;"  title="DescFor4,0">17</a></td><td><a href="javascript:;"  title="DescFor4,1">18</a></td><td><a href="javascript:;"  title="DescFor4,2">19</a></td><td><a href="javascript:;"  title="DescFor4,3">20</a></td></tr></table>'  # noqa
        self.assertEqual(self.sheet.to_html_table(), expected_no_skip)
        self.assertEqual(self.sheet.to_html_table(skip_labels=True),
                         expected_skip)

    def test_to_json(self):
        """Test export to JSON"""
        expected: str = '{"table": {"data": {"rows": {"R_0": {"columns": {"NL_C_0": {"excel": "1", "python_numpy": "1", "native": "1", "value": 1, "description": "DescFor0,0", "column_description": "HT_C_0"}, "NL_C_1": {"excel": "2", "python_numpy": "2", "native": "2", "value": 2, "description": "DescFor0,1", "column_description": "HT_C_1"}, "NL_C_2": {"excel": "3", "python_numpy": "3", "native": "3", "value": 3, "description": "DescFor0,2", "column_description": "HT_C_2"}, "NL_C_3": {"excel": "4", "python_numpy": "4", "native": "4", "value": 4, "description": "DescFor0,3", "column_description": "HT_C_3"}}, "row_description": "HT_R_0"}, "R_1": {"columns": {"NL_C_0": {"excel": "5", "python_numpy": "5", "native": "5", "value": 5, "description": "DescFor1,0", "column_description": "HT_C_0"}, "NL_C_1": {"excel": "6", "python_numpy": "6", "native": "6", "value": 6, "description": "DescFor1,1", "column_description": "HT_C_1"}, "NL_C_2": {"excel": "7", "python_numpy": "7", "native": "7", "value": 7, "description": "DescFor1,2", "column_description": "HT_C_2"}, "NL_C_3": {"excel": "8", "python_numpy": "8", "native": "8", "value": 8, "description": "DescFor1,3", "column_description": "HT_C_3"}}, "row_description": "HT_R_1"}, "R_2": {"columns": {"NL_C_0": {"excel": "9", "python_numpy": "9", "native": "9", "value": 9, "description": "DescFor2,0", "column_description": "HT_C_0"}, "NL_C_1": {"excel": "10", "python_numpy": "10", "native": "10", "value": 10, "description": "DescFor2,1", "column_description": "HT_C_1"}, "NL_C_2": {"excel": "11", "python_numpy": "11", "native": "11", "value": 11, "description": "DescFor2,2", "column_description": "HT_C_2"}, "NL_C_3": {"excel": "12", "python_numpy": "12", "native": "12", "value": 12, "description": "DescFor2,3", "column_description": "HT_C_3"}}, "row_description": "HT_R_2"}, "R_3": {"columns": {"NL_C_0": {"excel": "13", "python_numpy": "13", "native": "13", "value": 13, "description": "DescFor3,0", "column_description": "HT_C_0"}, "NL_C_1": {"excel": "14", "python_numpy": "14", "native": "14", "value": 14, "description": "DescFor3,1", "column_description": "HT_C_1"}, "NL_C_2": {"excel": "15", "python_numpy": "15", "native": "15", "value": 15, "description": "DescFor3,2", "column_description": "HT_C_2"}, "NL_C_3": {"excel": "16", "python_numpy": "16", "native": "16", "value": 16, "description": "DescFor3,3", "column_description": "HT_C_3"}}, "row_description": "HT_R_3"}, "R_4": {"columns": {"NL_C_0": {"excel": "17", "python_numpy": "17", "native": "17", "value": 17, "description": "DescFor4,0", "column_description": "HT_C_0"}, "NL_C_1": {"excel": "18", "python_numpy": "18", "native": "18", "value": 18, "description": "DescFor4,1", "column_description": "HT_C_1"}, "NL_C_2": {"excel": "19", "python_numpy": "19", "native": "19", "value": 19, "description": "DescFor4,2", "column_description": "HT_C_2"}, "NL_C_3": {"excel": "20", "python_numpy": "20", "native": "20", "value": 20, "description": "DescFor4,3", "column_description": "HT_C_3"}}, "row_description": "HT_R_4"}}}, "variables": {}, "row-labels": ["R_0", "R_1", "R_2", "R_3", "R_4"], "column-labels": ["NL_C_0", "NL_C_1", "NL_C_2", "NL_C_3"]}}'  # noqa
        expected_parsed_dict: dict = json.loads(expected)
        computed_parsed_dict: dict = json.loads(self.sheet.to_json())
        self.assertDictEqual(expected_parsed_dict, computed_parsed_dict)


class TestSerializationToArrays(unittest.TestCase):
    """Test to_numpy serializer."""

    def setUp(self) -> None:
        self.warnings = []
        self.nr_row = 20
        self.nr_col = 30
        self.rows_labels = [f"R_{r_i}" for r_i in range(self.nr_row)]
        self.columns_labels = [f"NL_C_{c_i}" for c_i in range(self.nr_col)]
        self.rows_help_text = [f"HT_R_{r_i}" for r_i in range(self.nr_row)]
        self.columns_help_text = [f"HT_C_{c_i}" for c_i in range(self.nr_col)]
        self.native_rows = [f"NL_R_{r_i}" for r_i in range(self.nr_row)]
        self.native_cols = [f"NL_C_{c_i}" for c_i in range(self.nr_col)]
        self.sheet = Spreadsheet.create_new_sheet(
            self.nr_row, self.nr_col, {
                'native': (
                    self.native_rows,
                    self.native_cols
                )
            },
            rows_labels=self.rows_labels,
            columns_labels=self.columns_labels,
            rows_help_text=self.rows_help_text,
            columns_help_text=self.columns_help_text,
            excel_append_row_labels=True,
            excel_append_column_labels=True,
            warning_logger=lambda message: self.warnings.append(message)
        )
        self.sheet_shape = (self.nr_row, self.nr_col)
        # Add some random values:
        self.inserted_rand_values: np.ndarray = \
            np.random.random(self.sheet_shape) * 1_000
        self.sheet.iloc[:, :] = self.inserted_rand_values

    def test_to_numpy(self):
        """Test exporting to NumPy"""
        self.assertTrue(np.allclose(self.sheet.to_numpy(),
                                    self.inserted_rand_values))

    def test_to_2d_list(self):
        """Test the serialization to 2D list"""
        # Check values
        computed_2d_list = self.sheet.to_2d_list(skip_labels=True)
        self.assertTrue(isinstance(computed_2d_list, list))
        self.assertTrue(
            np.allclose(np.array(computed_2d_list), self.inserted_rand_values)
            )
        # Check labels:
        corner = "yMq7W0bk"
        computed_2d_list = self.sheet.to_2d_list(skip_labels=False,
                                                 top_right_corner_text=corner)
        self.assertTrue(isinstance(computed_2d_list, list))
        self.assertTrue(computed_2d_list[0][0], corner)
        # Check row labels
        for row_idx, row_label in enumerate(
                self.sheet.cell_indices.rows_labels):
            self.assertEqual(computed_2d_list[row_idx + 1][0], row_label)
        # Check column labels
        for col_idx, col_label in enumerate(
                self.sheet.cell_indices.columns_labels):
            self.assertEqual(computed_2d_list[0][col_idx + 1], col_label)
        # Check values inside:
        computed = [arr[1:] for arr in computed_2d_list[1:]]
        self.assertTrue(
            np.allclose(np.array(computed), self.inserted_rand_values)
        )
        # Check the language export
        sheet = copy.deepcopy(self.sheet)
        sheet.iloc[0, 0] = sheet.iloc[0, 1] * sheet.iloc[1, 0]
        computed_2d_list = sheet.to_2d_list(skip_labels=True,
                                            language='excel')
        # Regression test
        self.assertEqual(sheet.iloc[0, 0].parse['excel'], '=C2*B3')
        # Test all the values inside
        for row_idx in range(sheet.shape[0]):
            for col_idx in range(sheet.shape[1]):
                computed = computed_2d_list[row_idx][col_idx]
                expected = sheet.iloc[row_idx, col_idx].parse['excel']
                self.assertEqual(expected, computed)

    def test_to_string_of_values(self):
        """Test the serialization to 2D list"""
        computed_2d_list = self.sheet.to_string_of_values()
        self.assertTrue(isinstance(computed_2d_list, str))
        evaluated_list = eval(computed_2d_list)
        self.assertTrue(
            np.allclose(np.array(evaluated_list), self.inserted_rand_values)
            )



from numbers import Number
from typing import Tuple, List, Union, Optional, Callable
import copy

from .cell import Cell
from .cell_indices import CellIndices, T_lg_col_row
from .cell_slice import CellSlice
from .spreadsheet_utils import _Location, _Functionality, _SheetVariables
from .serialization import Serialization

# ==== TYPES ====
# Type for the sheet (list of the list of the cells)
T_sheet = List[List[Cell]]
# Sheet cell value
T_cell_val = Union[Number, Cell]
# ===============


class Spreadsheet(Serialization):
    """Simple spreadsheet that keeps tracks of each operations in defined
        languages.

    Logic allows export sheets to Excel files (and see how each cell is
        computed), to the JSON strings with description of computation e. g.
        in native language. It also allows to reconstruct behaviours in native
        Python with Numpy.

    Attributes:
        self.cell_indices (CellIndices): Define indices and the shape of the
            spreadsheet.
        _sheet (T_sheet): Array holding actual sheet.
        iloc (_Location): To make cells accessible using
            obj.iloc[integer_index_x, integer_index_y]
        loc (_Location): To make cells accessible using
            obj.loc[nick_x, nick_y]
        fn (_Functionality): To make accessible shortcuts for functionality
        var (_SheetVariables): Variables in the sheet.
    """

    def __init__(self,
                 cell_indices: CellIndices,
                 warning_logger: Optional[Callable[[str], None]] = None):
        """Initialize the spreadsheet object

        Args:
            cell_indices (CellIndices): The definition of the shape and columns
                and rows labels, help texts and descriptors.
            warning_logger (Optional[Callable[[str], None]]): Function that
                logs the warnings (or None if skipped).
        """
        # Initialise functionality for serialization:
        super().__init__(warning_logger=warning_logger)

        self._cell_indices: CellIndices = copy.deepcopy(cell_indices)

        self._sheet: T_sheet = self._initialise_array()
        # To make cells accessible using obj.iloc[pos_x, pos_y]
        self.iloc: _Location = _Location(self, True)
        # To make cells accessible using obj.loc[nick_x, nick_y]
        self.loc: _Location = _Location(self, False)
        # To make accessible shortcuts for functionality
        self.fn: _Functionality = _Functionality(self)
        # Variables of the sheet
        self.var: _SheetVariables = _SheetVariables(self)

    @staticmethod
    def create_new_sheet(
            number_of_rows: int,
            number_of_columns: int,
            rows_columns: Optional[T_lg_col_row] = None,
            /, *,  # noqa E999
            rows_labels: List[str] = None,
            columns_labels: List[str] = None,
            rows_help_text: List[str] = None,
            columns_help_text: List[str] = None,
            excel_append_row_labels: bool = True,
            excel_append_column_labels: bool = True,
            warning_logger: Optional[Callable[[str], None]] = None
    ) -> 'Spreadsheet':
        """Direct way of creating instance.

        Args:
            number_of_rows (int): Number of rows.
            number_of_columns (int): Number of columns.
            rows_columns (T_lg_col_row): List of all row names and column names
                for each user defined language.
            rows_labels (List[str]): List of masks (aliases) for row
                names.
            columns_labels (List[str]): List of masks (aliases) for column
                names.
            rows_help_text (List[str]): List of help texts for each row.
            columns_help_text (List[str]): List of help texts for each column.
            excel_append_row_labels (bool): If True, one column is added
                on the beginning of the sheet as a offset for labels.
            excel_append_column_labels (bool): If True, one row is added
                on the beginning of the sheet as a offset for labels.
            warning_logger (Optional[Callable[[str], None]]): Function that
                logs the warnings (or None if skipped).

        Returns:
            Spreadsheet: New instance of spreadsheet.
        """
        class_index = CellIndices(
            number_of_rows,
            number_of_columns,
            rows_columns,
            rows_labels=rows_labels,
            columns_labels=columns_labels,
            rows_help_text=rows_help_text,
            columns_help_text=columns_help_text,
            excel_append_row_labels=excel_append_row_labels,
            excel_append_column_labels=excel_append_column_labels,
            warning_logger=warning_logger
        )
        return Spreadsheet(class_index, warning_logger)

    def _initialise_array(self) -> T_sheet:
        """Initialise the first empty spreadsheet array on the beginning.

        Returns:
            T_sheet: New empty spreadsheet.
        """
        array: T_sheet = []
        for row_idx in range(self.cell_indices.shape[0]):
            row: List[Cell] = []
            for col_idx in range(self.cell_indices.shape[1]):
                row.append(Cell(row_idx, col_idx,
                                cell_indices=self.cell_indices))
            array.append(row)
        return array

    def _set_item(self,
                  value: T_cell_val,
                  index_integer: Tuple[int, int] = None,
                  index_label: Tuple[object, object] = None) -> None:
        """Set the spreadsheet cell on the desired index to the new value.

        Args:
            value (T_cell_val): New value to be inserted.
            index_integer (Tuple[int, int]): Return the value on the integer
                index (row, column) inside spreadsheet (indexed from 0).
            index_label (Tuple[str, str]): Return the value on the index
                (row label, column label) inside spreadsheet.
        """
        if index_integer is not None and index_label is not None:
            raise ValueError("Only one of parameters 'index_integer' and"
                             "'index_label' has to be set!")
        if index_label is not None:
            _x = self.cell_indices.rows_labels.index(index_label[0])
            _y = self.cell_indices.columns_labels.index(index_label[1])
            index_integer = (_x, _y)
        if index_integer is not None:
            _x = index_integer[0]
            _y = index_integer[1]
            # If negative, take n-th item from the end
            if _x < 0:
                _x += self.shape[0]
            if _y < 0:
                _y += self.shape[1]
            if isinstance(value, Cell):
                if value.anchored:
                    _value = Cell.reference(value)
                else:
                    # Create a deep copy
                    _value = copy.deepcopy(value)
                    # Anchor it:
                    _value.coordinates = (_x, _y)
            else:
                _value = Cell(_x, _y,
                              value=value, cell_indices=self.cell_indices)
            self._sheet[_x][_y] = _value

    def _get_item(self,
                  index_integer: Tuple[int, int] = None,
                  index_label: Tuple[object, object] = None) -> Cell:
        """Get the cell on the particular index.

        Args:
            index_integer (Tuple[int, int]): Return the value on the integer
                index (row, column) inside spreadsheet (indexed from 0).
            index_label (Tuple[str, str]): Return the value on the index
                (row label, column label) inside spreadsheet.

        Returns:
            Cell: The Cell on the desired index.
        """
        if index_integer is not None and index_label is not None:
            raise ValueError("Only one of parameters 'index_integer' and"
                             "'index_label' has to be set!")
        if index_label is not None:
            _x = self.cell_indices.rows_labels.index(index_label[0])
            _y = self.cell_indices.columns_labels.index(index_label[1])
            index_integer = (_x, _y)
        if index_integer is not None:
            _x = index_integer[0]
            _y = index_integer[1]
            # If negative, take n-th item from the end
            if _x < 0:
                _x += self.shape[0]
            if _y < 0:
                _y += self.shape[1]
            return self._sheet[_x][_y]

    def _get_slice(self,
                   index_integer: Tuple[slice, slice],
                   index_label: Tuple[slice, slice]) -> CellSlice:
        """Get the values in the slice.

        Args:
            index_integer (Tuple[int, int]): The position of the slice in the
                spreadsheet. Mutually exclusive with parameter index_label
            index_label (Tuple[object, object]): The position of the slice
                in the spreadsheet. Mutually exclusive with parameter
                index_integer (only one can be set to not None).
        Returns:
            CellSlice: Slice of the cells (aggregate).
        """
        if index_integer is not None and index_label is not None:
            raise ValueError("Only one of parameters 'index_integer' and"
                             "'index_label' has to be set!")

        if index_label is not None:
            if isinstance(index_label[0], slice):
                # If the first index is slice
                _x_start = 0
                if index_label[0].start:
                    _x_start = self.cell_indices.rows_labels.index(
                        index_label[0].start)
                _x_end = self.shape[0]
                if index_label[0].stop:
                    _x_end = self.cell_indices.rows_labels.index(
                        index_label[0].stop)
                _x_step = 1
                if index_label[0].step:
                    _x_step = int(index_label[0].step)
            else:
                # If the first index is scalar
                _x_start = self.cell_indices.rows_labels.index(
                    index_label[0])
                _x_end = _x_start + 1
                _x_step = 1

            if isinstance(index_label[1], slice):
                # If the second index is slice
                _y_start = 0
                if index_label[1].start:
                    _y_start = self.cell_indices.columns_labels.index(
                        index_label[1].start)
                _y_end = self.shape[1]
                if index_label[1].stop:
                    _y_end = self.cell_indices.columns_labels.index(
                        index_label[1].stop)
                _y_step = 1
                if index_label[1].step:
                    _y_step = int(index_label[1].step)
            else:
                # If the first index is scalar
                _y_start = self.cell_indices.columns_labels.index(
                    index_label[1])
                _y_end = _y_start + 1
                _y_step = 1

        if index_integer is not None:
            if isinstance(index_integer[0], slice):
                # If the first index is slice
                _x_start = 0
                if index_integer[0].start:
                    _x_start = int(index_integer[0].start)
                    # Negative index starts from end
                    if _x_start < 0:
                        _x_start = self.shape[0] + _x_start
                _x_end = self.shape[0]
                if index_integer[0].stop:
                    _x_end = int(index_integer[0].stop)
                    # Negative index starts from end
                    if _x_end < 0:
                        _x_end = self.shape[0] + _x_end
                _x_step = 1
                if index_integer[0].step:
                    _x_step = int(index_integer[0].step)
            else:
                # If the first index is scalar
                _x_start = index_integer[0]
                _x_end = _x_start + 1
                _x_step = 1

            if isinstance(index_integer[1], slice):
                # If the second index is slice
                _y_start = 0
                if index_integer[1].start:
                    _y_start = int(index_integer[1].start)
                    # Negative index starts from end
                    if _y_start < 0:
                        _y_start = self.shape[1] + _y_start
                _y_end = self.shape[1]
                if index_integer[1].stop:
                    _y_end = int(index_integer[1].stop)
                    # Negative index starts from end
                    if _y_end < 0:
                        _y_end = self.shape[1] + _y_end
                _y_step = 1
                if index_integer[1].step:
                    _y_step = int(index_integer[1].step)
            else:
                # If the first index is scalar
                _y_start = index_integer[1]
                _y_end = _y_start + 1
                _y_step = 1

        # Create the CellSlice object
        cell_subset = []
        for x in range(_x_start, _x_end, _x_step):
            for y in range(_y_start, _y_end, _y_step):
                cell_subset.append(self.iloc[x, y])
        cell_slice: CellSlice = CellSlice((_x_start, _y_start),
                                          (_x_end - 1, _y_end - 1),
                                          cell_subset,
                                          self)
        return cell_slice

    def _set_slice(self,
                   value: T_cell_val,
                   index_integer: Tuple[int, int],
                   index_label: Tuple[object, object]) -> None:
        """Set the value of each cell in the slice

        Args:
            value (T_cell_val): New value to be set.
            index_integer (Tuple[int, int]): The position of the slice in the
                spreadsheet. Mutually exclusive with parameter index_label
            index_label (Tuple[object, object]): The position of the slice
                in the spreadsheet. Mutually exclusive with parameter
                index_integer (only one can be set to not None).
        """
        cell_slice: CellSlice = self._get_slice(index_integer, index_label)
        cell_slice.set(value)

    def expand(self,
               new_number_of_rows: int,
               new_number_of_columns: int,
               new_rows_columns: Optional[T_lg_col_row] = {},
               /, *,  # noqa E225
               new_rows_labels: List[str] = None,
               new_columns_labels: List[str] = None,
               new_rows_help_text: List[str] = None,
               new_columns_help_text: List[str] = None
               ):
        """Expand the size of the table.

        Args:
            new_number_of_rows (int): Number of rows to be added.
            new_number_of_columns (int): Number of columns to be added.
            new_rows_columns (T_lg_col_row): List of all row names and column
                names for each language to be added.
            new_rows_labels (List[str]): List of masks (aliases) for row
                names to be added.
            new_columns_labels (List[str]): List of masks (aliases) for
                column names to be added.
            new_rows_help_text (List[str]): List of help texts for each row to
                be added.
            new_columns_help_text (List[str]): List of help texts for each
                column to be added.
        """
        self.expand_using_cell_indices(
            self.cell_indices.expand_size(
                new_number_of_rows,
                new_number_of_columns,
                new_rows_columns,

                new_rows_labels=new_rows_labels,
                new_columns_labels=new_columns_labels,
                new_rows_help_text=new_rows_help_text,
                new_columns_help_text=new_columns_help_text
            )
        )

    def expand_using_cell_indices(self, cell_indices: CellIndices) -> None:
        """Resize the spreadsheet object to the greater size

        Args:
            cell_indices (CellIndices): The definition of the shape and columns
                and rows labels, help texts and descriptors.
        """
        shape_origin = self.shape
        self._cell_indices = copy.deepcopy(cell_indices)
        for row_idx in range(self.shape[0]):
            if row_idx >= shape_origin[0]:
                # Append wholly new rows
                row: List[Cell] = []
                for col in range(self.cell_indices.shape[1]):
                    row.append(Cell(cell_indices=self.cell_indices))
                self._sheet.append(row)
            else:
                # Expand columns:
                for col in range(self.cell_indices.shape[1] - shape_origin[1]):
                    self._sheet[row_idx].append(
                        Cell(cell_indices=self.cell_indices)
                    )
            for col_idx in range(self.shape[1]):
                # Has to refresh cell indices everywhere inside
                self.iloc[row_idx,
                          col_idx].cell_indices = self.cell_indices

    # ==== OVERRIDE ABSTRACT METHODS AND PROPERTIES OF SERIALIZATION CLASS ====
    @Serialization.shape.getter
    def shape(self) -> Tuple[int, int]:
        """Return the shape of the sheet in the NumPy logic.

        Returns:
            Tuple[int]: Number of rows, Number of columns
        """
        return self.cell_indices.shape

    @Serialization.cell_indices.getter
    def cell_indices(self) -> CellIndices:
        """Get the cell indices.

        Returns:
            CellIndices: Cell indices of the spreadsheet.
        """
        return self._cell_indices

    def _get_cell_at(self, row: int, column: int) -> 'Cell':
        """Get the particular cell on the (row, column) position.

        Returns:
            Cell: The call on given position.
        """
        return self.iloc[row, column]

    def _get_variables(self) -> '_SheetVariables':
        """Return the sheet variables as _SheetVariables object.

        Returns:
            _SheetVariables: Sheet variables.
        """
        return self.var
    # =========================================================================



from typing import Dict, Union, Optional
import re
import copy
from numbers import Number

from .cell import Cell
from .cell_type import CellType

# ========== File with the functionality for internal purposes only ===========


class _Location(object):
    """Private class that enables indexing and slicing of values using
        obj.loc[col, row] or obj.iloc[col_idx, row_idx] logic.

    Attributes:
        spreadsheet (Spreadsheet): Reference to spreadsheet instance.
        by_integer (bool): If True, indices are computed using integer value,
            if False, labels (aliases, typically string) are used.
    """

    def __init__(self,
                 spreadsheet,
                 by_integer: bool):
        """Initialise location

        Args:
            spreadsheet (Spreadsheet): Reference to spreadsheet instance.
            by_integer (bool): If True, indices are computed using integer
            value, if False, labels (aliases, typically string) are used.
        """
        self.spreadsheet = spreadsheet
        self.by_integer: str = by_integer

    def __setitem__(self, index, val):
        """Set item selected as index or as a slice.

        Args:
            index: standard tuple of indices or tuple of slices.
            val: the value to be set.
        """
        has_slice = isinstance(index[0], slice) or isinstance(index[1], slice)
        if not has_slice:
            if self.by_integer:
                self.spreadsheet._set_item(val, index, None)
            else:
                self.spreadsheet._set_item(val, None, index)
        else:
            if self.by_integer:
                return self.spreadsheet._set_slice(val, index, None)
            else:
                return self.spreadsheet._set_slice(val, None, index)

    def __getitem__(self, index):
        """Get the item from the slice or single value.

        Args:
            index: standard tuple of indices or tuple of slices.
        """
        has_slice = isinstance(index[0], slice) or isinstance(index[1], slice)
        if not has_slice:
            if self.by_integer:
                return self.spreadsheet._get_item(index, None)
            else:
                return self.spreadsheet._get_item(None, index)
        else:
            if self.by_integer:
                return self.spreadsheet._get_slice(index, None)
            else:
                return self.spreadsheet._get_slice(None, index)


class _Functionality(object):
    """Class encapsulating some shortcuts for functionality.

    Attributes:
        spreadsheet (Spreadsheet): Reference to spreadsheet instance.
    """

    def __init__(self, spreadsheet):
        """
        Args:
            spreadsheet (Spreadsheet): Reference to spreadsheet instance.
        """
        self.spreadsheet = spreadsheet

    def const(self, value: Number) -> Cell:
        """Create the constant for computation (un-anchored cell).

        Args:
            value (Number): Some constant value.

        Returns:
            Cell: un-anchored cell with constant value.
        """
        return Cell(value=value,
                    cell_indices=self.spreadsheet.cell_indices)

    @staticmethod
    def raw(value: Cell, words: Dict[str, str]) -> Cell:
        """Add the raw statement and use the value of input cell.

        Args:
            value (Cell): Input cell that defines value and type of output.
            words (Dict[str, str]): Word for each language (language is a key,
                word is a value)

        Warnings:
            Do not use this feature unless you really have to.

        Returns:
            Cell: Expression with defined word
        """
        return Cell.raw(value, words)

    @staticmethod
    def brackets(body: Cell) -> Cell:
        """Shortcut for adding bracket around body.

        Args:
            body (Cell): The body of the expression.

        Returns:
            Cell: Expression with brackets
        """
        return Cell.brackets(body)

    @staticmethod
    def ln(value: Cell) -> Cell:
        """Natural logarithm of the value.

        Args:
            value (Cell): The input value for computation.

        Returns:
            Cell: Natural logarithm of the input value.
        """
        return Cell.logarithm(value)

    @staticmethod
    def exp(value: Cell) -> Cell:
        """Exponential function of the value (e^value).

        Args:
            value (Cell): The input value for computation.

        Returns:
            Cell: Exponential function of the input value.
        """
        return Cell.exponential(value)

    @staticmethod
    def floor(value: Cell) -> Cell:
        """Floor function of the value.

        Args:
            value (Cell): The input value for computation.

        Returns:
            Cell: Floor function of the input value.
        """
        return Cell.floor(value)

    @staticmethod
    def ceil(value: Cell) -> Cell:
        """Ceiling function of the value.

        Args:
            value (Cell): The input value for computation.

        Returns:
            Cell: Ceiling function of the input value.
        """
        return Cell.ceil(value)

    @staticmethod
    def round(value: Cell) -> Cell:
        """Round the numeric value.

        Args:
            value (Cell): The input value for computation.

        Returns:
            Cell: Round of the input value.
        """
        return Cell.round(value)

    @staticmethod
    def abs(value: Cell) -> Cell:
        """Absolute value of the input.

        Args:
            value (Cell): The input value for computation.

        Returns:
            Cell: Absolute value of the input value.
        """
        return Cell.abs(value)

    @staticmethod
    def sqrt(value: Cell) -> Cell:
        """Square root of the input.

        Args:
            value (Cell): The input value for computation.

        Returns:
            Cell: Square root of the input value.
        """
        return Cell.sqrt(value)

    @staticmethod
    def sign(value: Cell) -> Cell:
        """Signum function of the input.

        Args:
            value (Cell): The input value for computation.

        Returns:
            Cell: Signum function value of the input value.
        """
        return Cell.signum(value)

    @staticmethod
    def neg(value: Cell) -> Cell:
        """Logical negation of the input.

        Args:
            value (Cell): The input value for computation.

        Returns:
            Cell: logical negation of the input value.
        """
        return Cell.logicalNegation(value)

    @staticmethod
    def conditional(condition: Cell,
                    consequent: Cell,
                    alternative: Cell, /) -> Cell:  # noqa E225
        """Conditional statement (standard if-then-else statement).

        Evaluate the value of the condition, if it is true, take the value
            of the consequent, if not take the value of the alternative.

        Args:
            condition (Cell): Cell defining the condition (boolean value)
            consequent (Cell): What is pass if the condition is true (the
                part right after the if)
            alternative (Cell): What is pass if the condition is false (the
                part right after the else)

        Returns:
            Cell: The if-then-else conditional statement and correctly
                chosen value.
        """
        return Cell.conditional(condition, consequent, alternative)

    def offset(self,
               reference: Cell,
               row_skip: Cell,
               column_skip: Cell, /) -> Cell:  # noqa E225
        """Return the cell with value computed as offset from reference
            cell plus row_skip rows and column_skip columns.

        Args:
            reference (Cell): Reference cell from that the position is
                computed.
            row_skip (Cell): How many rows (down) should be skipped.
            column_skip (Cell): How many columns (left) should be skipped.

        Returns:
            Cell: Cell with offset operation and value from the cell on the
                referential position.

        Raises:
            ValueError: If the reference cell is not anchored.
        """
        # Test if the reference cell is anchored
        if not reference.anchored:
            raise ValueError("Reference cell must be anchored!")
        # Find the target cell
        target = self.spreadsheet.iloc[
            reference.row + row_skip.value,
            reference.column + column_skip.value
        ]
        return Cell.offset(reference, row_skip, column_skip, target=target)


class _SheetVariables(object):
    """Encapsulate the sheet variables.

    Attributes:
        _variables (Dict[str, object]): Dictionary with variables, key is the
            name of the variable and value is the actual value assigned to the
            variable.
        spreadsheet (Spreadsheet): Reference to spreadsheet instance.
    """
    def __init__(self, spreadsheet):
        """
        Args:
            spreadsheet (Spreadsheet): Reference to spreadsheet instance.
        """
        self._variables: Dict[str, Dict[str, object]] = {}
        self.spreadsheet = spreadsheet

    def set_variable(self,
                     name: str,
                     value: Union[str, Number],
                     description: Optional[str] = None,
                     exclude_description_update: bool = False) -> None:
        """Check the name consistency and set the variable.

        If the variable with given name is already in the dictionary, it is
            rewritten with a new value.

        Args:
            name (str): Unique name for the variable. Has to be lower case
                alphanumeric value maximally with underscore symbols.
            value (Union[str, Number]): A value to be set.
            description (Optional[str]): Description of the variable
            exclude_description_update (bool): If true, description is not
                updated

        Raises:
            ValueError: If the structure of the name does not match the
                required pattern (lowercase alphanumeric and underscores only).

        Warning:
            The function str(VAL) is called on the value (it has to be stored
                as a string) when the result is exported to Excel!
        """
        # Check the consistency of name for variable
        if not isinstance(name, str):
            raise ValueError("Name has to be a string!")
        if ' ' in name:
            raise ValueError("Space is not allowed in the name of variable!")
        elif name.lower() != name:
            raise ValueError("Name of the variable has to be lowercase!")
        if re.match(r'^[a-z0-9_]+$', name):
            pass
        else:
            raise ValueError("Name of variable has to be alphanumeric value "
                             "maximally with underscores!")

        # Use the existing description if required
        if name in self._variables.keys():
            if exclude_description_update:
                description = self._variables[name]["description"]

        # Add the value to the dictionary:
        self._variables[name] = {
            "value": value,
            "description": description
        }

    @property
    def variables_dict(self) -> Dict[str, Dict[str, object]]:
        """Return the variables as a read-only dictionary with keys as a name
            of variable and value as actual value of the variable.

        Returns:
            Dict[str, Dict[str, object]]: Read-only dictionary with names of
                variables (keys) and their values and description (value).
                Example: 'key' -> {'value': VALUE, "description": DESCRIPTION}
        """
        return copy.deepcopy(self._variables)

    def variable_exist(self, name: str) -> bool:
        """Returns true if the variable exist in the dictionary.

        Args:
            name (str): Unique name for the variable. Has to be lower case
                alphanumeric value maximally with underscore symbols.
        Return:
            bool: True if variable exists in the system, False otherwise.
        """
        return str(name) in self._variables.keys()

    def get_variable(self, name: str, /) -> Cell:  # noqa E225
        """Get variable value as a new Cell by its name.

        Args:
            name (str): Name of the variable
        """
        if self.variable_exist(name):
            return Cell.variable(
                Cell(None,
                     None,
                     variable_name=name,
                     value=self._variables[name]['value'],
                     is_variable=True,
                     cell_type=CellType.computational,
                     cell_indices=self.spreadsheet.cell_indices
                     )
            )
        else:
            raise ValueError(f"Variable with name {name} does not exist in "
                             f"the system!")

    def __getitem__(self, item):
        """Overloads [item] operator getter"""
        return self.get_variable(item)

    def __setitem__(self, key, value):
        """Overloads [item] operator setter"""
        self.set_variable(key, value, exclude_description_update=True)

    @property
    def empty(self) -> bool:
        """Return True if the variable set is empty, False otherwise.

        Returns:
            bool: Return True if the variable set is empty, False otherwise.
        """
        return len(self._variables) == 0



from enum import Enum


class CellType(Enum):
    """Reflect the cell type in the table, there are just two types one is the
        cell holding some constant and another one is the cell doing some
        computations.
    """
    # Type of cell that contains only constant
    value_only = "value_only"
    # Type of cell that does computations
    computational = "computational"



from .spreadsheet import Spreadsheet  # noqa
from .cell import Cell  # noqa
from .cell_indices import CellIndices  # noqa
from .cell_slice import CellSlice  # noqa
from .grammars import GRAMMARS  # noqa
from .grammar_utils import GrammarUtils  # noqa

__version__ = "0.1.11"
__status__ = "Production"
